import torch.nn.functional as F
from torch.nn.parameter import Parameter
import numpy.random as npr
from collections import OrderedDict
import random
from torch.utils.data import Dataset
import torchvision.transforms as transforms
from torch.utils.data import sampler
import lmdb
import six
import sys
from PIL import Image
import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
import collections
import cv2
import os, shutil
from ttcv.basic.basic_objects import BasDetectObj


class MoranTextRecognizer(BasDetectObj):
    def get_model(self):
        self.cuda_flag = False
        if torch.cuda.is_available():
            self.cuda_flag = True
            self.model = MORAN(1, len(self.args.alphabet.split(':')), 256, 32, 100, BidirDecoder=True, CUDA=self.cuda_flag)
            self.model = self.model.cuda()
        else:
            self.model = MORAN(1, len(self.args.alphabet.split(':')), 256, 32, 100, BidirDecoder=True,
                          inputDataType='torch.FloatTensor', CUDA=self.cuda_flag)
        # ----------- >>>> Load pretrained model
        if self.cuda_flag:
            state_dict = torch.load(self.args.text_recognizer_checkpoint)
        else:
            state_dict = torch.load(self.args.text_recognizer_checkpoint, map_location='cpu')
        MORAN_state_dict_rename = OrderedDict()
        for k, v in state_dict.items():
            name = k.replace("module.", "")  # remove `module.`
            MORAN_state_dict_rename[name] = v
        self.model.load_state_dict(MORAN_state_dict_rename)

        for p in self.model.parameters():
            p.requires_grad = False
        self.model.eval()

        self.converter = strLabelConverterForAttention(self.args.alphabet, ':')
        self.transformer = resizeNormalize((100, 32))

    def get_probs_string(self, preds_in, length, flip=False):
        preds = F.softmax(preds_in, dim=1)
        probs, preds = preds.max(1)
        probs = probs.cpu().numpy()

        sim_preds = self.converter.decode(preds.data, length.data)
        text_len = sim_preds.find('$')

        if not any(el.isdigit() for el in sim_preds):  # no digit in text
            sim_preds = ''
            text_len = 0

        if text_len == 0:
            return '', text_len, [],0.0

        sim_preds = sim_preds[:text_len]
        probs = np.abs(probs[:text_len])
        prob = probs.mean()

        if flip: return sim_preds[::-1], text_len, probs[::-1], prob
        return sim_preds, text_len, probs, prob


    def predict_arrays(self, im_in, label='unlabeled'):

        if self.args.show_steps:
            print('-' * 80)
            print('image_path \tpredict_label \treverse_label')
            print('-' * 80)

        out_steps_dir = os.path.join(self.args.root_dir, self.args.out_steps)
        if self.args.show_steps:
            out_steps_dir = os.path.join(self.args.root_dir, self.args.out_steps)
            if os.path.exists(out_steps_dir): shutil.rmtree(out_steps_dir)
            os.makedirs(out_steps_dir)

        im_list = [im_in, np.rot90(im_in, axes=(0, 1)), np.rot90(im_in, axes=(1, 0))]

        best_pred = None
        pred_ind = -1
        pred_prob = -1

        texts = []
        for i in range(3):
            image = Image.fromarray(im_list[i]).convert('L')
            image = self.transformer(image)

            if self.cuda_flag:
                image = image.cuda()
            image = image.view(1, *image.size())
            image = Variable(image)
            text = torch.LongTensor(1 * 5)
            length = torch.IntTensor(1)
            text = Variable(text)
            length = Variable(length)

            max_iter = 20
            t, l = self.converter.encode('0' * max_iter)
            loadData(text, t)
            loadData(length, l)
            output = self.model(image, length, text, text, test=True, debug=True)

            preds, preds_reverse = output[0]
            demo = output[1]

            sim_preds, text_len, probs, prob = self.get_probs_string(preds, length)
            sim_preds_reverse, text_len_reverse, probs_reverse, prob_reverse = self.get_probs_string(preds_reverse, length, flip=True)

            if self.args.show_steps:
                print('%s \t\t%s \t\t%s' % (label, sim_preds, sim_preds_reverse))
                cv2.imwrite(os.path.join(out_steps_dir, '%s_%s_%s_%d.png'%(label,sim_preds, sim_preds_reverse, i)), demo)

            if max(text_len,text_len_reverse) == 0: continue

            len_diff = abs(text_len - text_len_reverse)
            len_avg = (text_len + text_len_reverse)/2
            alpha = (len_avg - len_diff) / len_avg
            if len_diff != 0:
                baselen_diff = abs(text_len - self.args.text_baselen)
                baselen_diff_reverse = abs(text_len_reverse - self.args.text_baselen)
                if baselen_diff < baselen_diff_reverse:
                    text1 = sim_preds
                    prob1 = prob*alpha
                elif baselen_diff > baselen_diff_reverse:
                    text1 = sim_preds_reverse
                    prob1 = prob_reverse*alpha

                if prob1 > pred_prob:
                    best_pred = text1
                    pred_prob = prob1
            else:
                new_preds = ''
                new_probs = []
                for prob, prob_r, sim, sim_r in zip(probs, probs_reverse, sim_preds, sim_preds_reverse):
                    if prob > prob_r:
                        new_preds += sim
                        new_probs.append(prob)
                    else:
                        new_preds += sim_r
                        new_probs.append(prob_r)
                prob1 = np.mean(np.array(new_probs))
                if prob1 > pred_prob:
                    best_pred = new_preds
                    pred_prob = prob1
        return best_pred

class MORAN(nn.Module):

    def __init__(self, nc, nclass, nh, targetH, targetW, BidirDecoder=False,
    	inputDataType='torch.cuda.FloatTensor', maxBatch=256, CUDA=True):
        super(MORAN, self).__init__()
        self.MORN = MORN(nc, targetH, targetW, inputDataType, maxBatch, CUDA)
        self.ASRN = ASRN(targetH, nc, nclass, nh, BidirDecoder, CUDA)

    def forward(self, x, length, text, text_rev, test=False, debug=False):
        if debug:
            x_rectified, demo = self.MORN(x, test, debug=debug)
            preds = self.ASRN(x_rectified, length, text, text_rev, test)
            return preds, demo
        else:
            x_rectified = self.MORN(x, test, debug=debug)
            preds = self.ASRN(x_rectified, length, text, text_rev, test)
            return preds




class MORN(nn.Module):
    def __init__(self, nc, targetH, targetW, inputDataType='torch.cuda.FloatTensor', maxBatch=256, CUDA=True):
        super(MORN, self).__init__()
        self.targetH = targetH
        self.targetW = targetW
        self.inputDataType = inputDataType
        self.maxBatch = maxBatch
        self.cuda = CUDA

        self.cnn = nn.Sequential(
            nn.MaxPool2d(2, 2),
            nn.Conv2d(nc, 64, 3, 1, 1), nn.BatchNorm2d(64), nn.ReLU(True), nn.MaxPool2d(2, 2),
            nn.Conv2d(64, 128, 3, 1, 1), nn.BatchNorm2d(128), nn.ReLU(True), nn.MaxPool2d(2, 2),
            nn.Conv2d(128, 64, 3, 1, 1), nn.BatchNorm2d(64), nn.ReLU(True),
            nn.Conv2d(64, 16, 3, 1, 1), nn.BatchNorm2d(16), nn.ReLU(True),
            nn.Conv2d(16, 1, 3, 1, 1), nn.BatchNorm2d(1)
        )

        self.pool = nn.MaxPool2d(2, 1)

        h_list = np.arange(self.targetH) * 2. / (self.targetH - 1) - 1
        w_list = np.arange(self.targetW) * 2. / (self.targetW - 1) - 1

        grid = np.meshgrid(
            w_list,
            h_list,
            indexing='ij'
        )
        grid = np.stack(grid, axis=-1)
        grid = np.transpose(grid, (1, 0, 2))
        grid = np.expand_dims(grid, 0)
        grid = np.tile(grid, [maxBatch, 1, 1, 1])
        grid = torch.from_numpy(grid).type(self.inputDataType)
        if self.cuda:
            grid = grid.cuda()

        self.grid = Variable(grid, requires_grad=False)
        self.grid_x = self.grid[:, :, :, 0].unsqueeze(3)
        self.grid_y = self.grid[:, :, :, 1].unsqueeze(3)

    def forward(self, x, test, enhance=1, debug=False):

        if not test and np.random.random() > 0.5:
            return nn.functional.upsample(x, size=(self.targetH, self.targetW), mode='bilinear')
        if not test:
            enhance = 0

        assert x.size(0) <= self.maxBatch
        assert x.data.type() == self.inputDataType

        grid = self.grid[:x.size(0)]
        grid_x = self.grid_x[:x.size(0)]
        grid_y = self.grid_y[:x.size(0)]
        x_small = nn.functional.upsample(x, size=(self.targetH, self.targetW), mode='bilinear')

        offsets = self.cnn(x_small)
        offsets_posi = nn.functional.relu(offsets, inplace=False)
        offsets_nega = nn.functional.relu(-offsets, inplace=False)
        offsets_pool = self.pool(offsets_posi) - self.pool(offsets_nega)

        offsets_grid = nn.functional.grid_sample(offsets_pool, grid)
        offsets_grid = offsets_grid.permute(0, 2, 3, 1).contiguous()
        offsets_x = torch.cat([grid_x, grid_y + offsets_grid], 3)
        x_rectified = nn.functional.grid_sample(x, offsets_x)

        for iteration in range(enhance):
            offsets = self.cnn(x_rectified)

            offsets_posi = nn.functional.relu(offsets, inplace=False)
            offsets_nega = nn.functional.relu(-offsets, inplace=False)
            offsets_pool = self.pool(offsets_posi) - self.pool(offsets_nega)

            offsets_grid += nn.functional.grid_sample(offsets_pool, grid).permute(0, 2, 3, 1).contiguous()
            offsets_x = torch.cat([grid_x, grid_y + offsets_grid], 3)
            x_rectified = nn.functional.grid_sample(x, offsets_x)

        if debug:

            offsets_mean = torch.mean(offsets_grid.view(x.size(0), -1), 1)
            offsets_max, _ = torch.max(offsets_grid.view(x.size(0), -1), 1)
            offsets_min, _ = torch.min(offsets_grid.view(x.size(0), -1), 1)

            import matplotlib.pyplot as plt
            from colour import Color
            from torchvision import transforms
            import cv2

            alpha = 0.7
            density_range = 256
            color_map = np.empty([self.targetH, self.targetW, 3], dtype=int)
            cmap = plt.get_cmap("rainbow")
            blue = Color("blue")
            hex_colors = list(blue.range_to(Color("red"), density_range))
            rgb_colors = [[rgb * 255 for rgb in color.rgb] for color in hex_colors][::-1]
            to_pil_image = transforms.ToPILImage()

            for i in range(x.size(0)):

                img_small = x_small[i].data.cpu().mul_(0.5).add_(0.5)
                img = to_pil_image(img_small)
                img = np.array(img)
                if len(img.shape) == 2:
                    img = cv2.merge([img.copy()] * 3)
                img_copy = img.copy()

                v_max = offsets_max.data[i]
                v_min = offsets_min.data[i]
                img_offsets = (offsets_grid[i]).view(1, self.targetH, self.targetW).data.cpu().add_(-v_min).mul_(
                    1. / (v_max - v_min))
                img_offsets = to_pil_image(img_offsets)
                img_offsets = np.array(img_offsets)
                color_map = np.empty([self.targetH, self.targetW, 3], dtype=int)
                for h_i in range(self.targetH):
                    for w_i in range(self.targetW):
                        color_map[h_i][w_i] = rgb_colors[int(img_offsets[h_i, w_i] / 256. * density_range)]
                color_map = color_map.astype(np.uint8)
                cv2.addWeighted(color_map, alpha, img_copy, 1 - alpha, 0, img_copy)

                img_processed = x_rectified[i].data.cpu().mul_(0.5).add_(0.5)
                img_processed = to_pil_image(img_processed)
                img_processed = np.array(img_processed)
                if len(img_processed.shape) == 2:
                    img_processed = cv2.merge([img_processed.copy()] * 3)

                total_img = np.ones([self.targetH, self.targetW * 3 + 10, 3], dtype=int) * 255
                total_img[0:self.targetH, 0:self.targetW] = img
                total_img[0:self.targetH, self.targetW + 5:2 * self.targetW + 5] = img_copy
                total_img[0:self.targetH, self.targetW * 2 + 10:3 * self.targetW + 10] = img_processed
                total_img = cv2.resize(total_img.astype(np.uint8), (300, 50))
                # cv2.imshow("Input_Offsets_Output", total_img)
                # cv2.waitKey()

            return x_rectified, total_img

        return x_rectified


class BidirectionalLSTM(nn.Module):

    def __init__(self, nIn, nHidden, nOut):
        super(BidirectionalLSTM, self).__init__()

        self.rnn = nn.LSTM(nIn, nHidden, bidirectional=True, dropout=0.3)
        self.embedding = nn.Linear(nHidden * 2, nOut)

    def forward(self, input):
        recurrent, _ = self.rnn(input)
        T, b, h = recurrent.size()
        t_rec = recurrent.view(T * b, h)

        output = self.embedding(t_rec)  # [T * b, nOut]
        output = output.view(T, b, -1)

        return output


class AttentionCell(nn.Module):
    def __init__(self, input_size, hidden_size, num_embeddings=128, CUDA=True):
        super(AttentionCell, self).__init__()
        self.i2h = nn.Linear(input_size, hidden_size, bias=False)
        self.h2h = nn.Linear(hidden_size, hidden_size)
        self.score = nn.Linear(hidden_size, 1, bias=False)
        self.rnn = nn.GRUCell(input_size + num_embeddings, hidden_size)
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.num_embeddings = num_embeddings
        self.fracPickup = fracPickup(CUDA=CUDA)

    def forward(self, prev_hidden, feats, cur_embeddings, test=False):
        nT = feats.size(0)
        nB = feats.size(1)
        nC = feats.size(2)
        hidden_size = self.hidden_size

        feats_proj = self.i2h(feats.view(-1, nC))
        prev_hidden_proj = self.h2h(prev_hidden).view(1, nB, hidden_size).expand(nT, nB, hidden_size).contiguous().view(
            -1, hidden_size)
        emition = self.score(F.tanh(feats_proj + prev_hidden_proj).view(-1, hidden_size)).view(nT, nB)

        alpha = F.softmax(emition, 0)  # nT * nB

        if not test:
            alpha_fp = self.fracPickup(alpha.transpose(0, 1).contiguous().unsqueeze(1).unsqueeze(2)).squeeze()
            context = (feats * alpha_fp.transpose(0, 1).contiguous().view(nT, nB, 1).expand(nT, nB, nC)).sum(0).squeeze(
                0)  # nB * nC
            if len(context.size()) == 1:
                context = context.unsqueeze(0)
            context = torch.cat([context, cur_embeddings], 1)
            cur_hidden = self.rnn(context, prev_hidden)
            return cur_hidden, alpha_fp
        else:
            context = (feats * alpha.view(nT, nB, 1).expand(nT, nB, nC)).sum(0).squeeze(0)  # nB * nC
            if len(context.size()) == 1:
                context = context.unsqueeze(0)
            context = torch.cat([context, cur_embeddings], 1)
            cur_hidden = self.rnn(context, prev_hidden)
            return cur_hidden, alpha


class Attention(nn.Module):
    def __init__(self, input_size, hidden_size, num_classes, num_embeddings=128, CUDA=True):
        super(Attention, self).__init__()
        self.attention_cell = AttentionCell(input_size, hidden_size, num_embeddings, CUDA=CUDA)
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.generator = nn.Linear(hidden_size, num_classes)
        self.char_embeddings = Parameter(torch.randn(num_classes + 1, num_embeddings))
        self.num_embeddings = num_embeddings
        self.num_classes = num_classes
        self.cuda = CUDA

    # targets is nT * nB
    def forward(self, feats, text_length, text, test=False):

        nT = feats.size(0)
        nB = feats.size(1)
        nC = feats.size(2)
        hidden_size = self.hidden_size
        input_size = self.input_size
        assert (input_size == nC)
        assert (nB == text_length.numel())

        num_steps = text_length.data.max()
        num_labels = text_length.data.sum()

        if not test:

            targets = torch.zeros(nB, num_steps + 1).long()
            if self.cuda:
                targets = targets.cuda()
            start_id = 0

            for i in range(nB):
                targets[i][1:1 + text_length.data[i]] = text.data[start_id:start_id + text_length.data[i]] + 1
                start_id = start_id + text_length.data[i]
            targets = Variable(targets.transpose(0, 1).contiguous())

            output_hiddens = Variable(torch.zeros(num_steps, nB, hidden_size).type_as(feats.data))
            hidden = Variable(torch.zeros(nB, hidden_size).type_as(feats.data))

            for i in range(num_steps):
                cur_embeddings = self.char_embeddings.index_select(0, targets[i])
                hidden, alpha = self.attention_cell(hidden, feats, cur_embeddings, test)
                output_hiddens[i] = hidden

            new_hiddens = Variable(torch.zeros(num_labels, hidden_size).type_as(feats.data))
            b = 0
            start = 0

            for length in text_length.data:
                new_hiddens[start:start + length] = output_hiddens[0:length, b, :]
                start = start + length
                b = b + 1

            probs = self.generator(new_hiddens)
            return probs

        else:

            hidden = Variable(torch.zeros(nB, hidden_size).type_as(feats.data))
            targets_temp = Variable(torch.zeros(nB).long().contiguous())
            probs = Variable(torch.zeros(nB * num_steps, self.num_classes))
            if self.cuda:
                targets_temp = targets_temp.cuda()
                probs = probs.cuda()

            for i in range(num_steps):
                cur_embeddings = self.char_embeddings.index_select(0, targets_temp)
                hidden, alpha = self.attention_cell(hidden, feats, cur_embeddings, test)
                hidden2class = self.generator(hidden)
                probs[i * nB:(i + 1) * nB] = hidden2class
                _, targets_temp = hidden2class.max(1)
                targets_temp += 1

            probs = probs.view(num_steps, nB, self.num_classes).permute(1, 0, 2).contiguous()
            probs = probs.view(-1, self.num_classes).contiguous()
            probs_res = Variable(torch.zeros(num_labels, self.num_classes).type_as(feats.data))
            b = 0
            start = 0

            for length in text_length.data:
                probs_res[start:start + length] = probs[b * num_steps:b * num_steps + length]
                start = start + length
                b = b + 1

            return probs_res


class Residual_block(nn.Module):
    def __init__(self, c_in, c_out, stride):
        super(Residual_block, self).__init__()
        self.downsample = None
        flag = False
        if isinstance(stride, tuple):
            if stride[0] > 1:
                self.downsample = nn.Sequential(nn.Conv2d(c_in, c_out, 3, stride, 1),
                                                nn.BatchNorm2d(c_out, momentum=0.01))
                flag = True
        else:
            if stride > 1:
                self.downsample = nn.Sequential(nn.Conv2d(c_in, c_out, 3, stride, 1),
                                                nn.BatchNorm2d(c_out, momentum=0.01))
                flag = True
        if flag:
            self.conv1 = nn.Sequential(nn.Conv2d(c_in, c_out, 3, stride, 1),
                                       nn.BatchNorm2d(c_out, momentum=0.01))
        else:
            self.conv1 = nn.Sequential(nn.Conv2d(c_in, c_out, 1, stride, 0),
                                       nn.BatchNorm2d(c_out, momentum=0.01))
        self.conv2 = nn.Sequential(nn.Conv2d(c_out, c_out, 3, 1, 1),
                                   nn.BatchNorm2d(c_out, momentum=0.01))
        self.relu = nn.ReLU()

    def forward(self, x):
        residual = x
        conv1 = self.conv1(x)
        conv2 = self.conv2(conv1)
        if self.downsample is not None:
            residual = self.downsample(residual)
        return self.relu(residual + conv2)


class ResNet(nn.Module):
    def __init__(self, c_in):
        super(ResNet, self).__init__()
        self.block0 = nn.Sequential(nn.Conv2d(c_in, 32, 3, 1, 1), nn.BatchNorm2d(32, momentum=0.01))
        self.block1 = self._make_layer(32, 32, 2, 3)
        self.block2 = self._make_layer(32, 64, 2, 4)
        self.block3 = self._make_layer(64, 128, (2, 1), 6)
        self.block4 = self._make_layer(128, 256, (2, 1), 6)
        self.block5 = self._make_layer(256, 512, (2, 1), 3)

    def _make_layer(self, c_in, c_out, stride, repeat=3):
        layers = []
        layers.append(Residual_block(c_in, c_out, stride))
        for i in range(repeat - 1):
            layers.append(Residual_block(c_out, c_out, 1))
        return nn.Sequential(*layers)

    def forward(self, x):
        block0 = self.block0(x)
        block1 = self.block1(block0)
        block2 = self.block2(block1)
        block3 = self.block3(block2)
        block4 = self.block4(block3)
        block5 = self.block5(block4)
        return block5


class ASRN(nn.Module):

    def __init__(self, imgH, nc, nclass, nh, BidirDecoder=False, CUDA=True):
        super(ASRN, self).__init__()
        assert imgH % 16 == 0, 'imgH must be a multiple of 16'

        self.cnn = ResNet(nc)

        self.rnn = nn.Sequential(
            BidirectionalLSTM(512, nh, nh),
            BidirectionalLSTM(nh, nh, nh),
        )

        self.BidirDecoder = BidirDecoder
        if self.BidirDecoder:
            self.attentionL2R = Attention(nh, nh, nclass, 256, CUDA=CUDA)
            self.attentionR2L = Attention(nh, nh, nclass, 256, CUDA=CUDA)
        else:
            self.attention = Attention(nh, nh, nclass, 256, CUDA=CUDA)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal(m.weight, mode='fan_out', a=0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant(m.weight, 1)
                nn.init.constant(m.bias, 0)

    def forward(self, input, length, text, text_rev, test=False):
        # conv features
        conv = self.cnn(input)

        b, c, h, w = conv.size()
        assert h == 1, "the height of conv must be 1"
        conv = conv.squeeze(2)
        conv = conv.permute(2, 0, 1).contiguous()  # [w, b, c]

        # rnn features
        rnn = self.rnn(conv)

        if self.BidirDecoder:
            outputL2R = self.attentionL2R(rnn, length, text, test)
            outputR2L = self.attentionR2L(rnn, length, text_rev, test)
            return outputL2R, outputR2L
        else:
            output = self.attention(rnn, length, text, test)
            return output


class fracPickup(nn.Module):

    def __init__(self, CUDA=True):
        super(fracPickup, self).__init__()
        self.cuda = CUDA

    def forward(self, x):
        x_shape = x.size()
        assert len(x_shape) == 4
        assert x_shape[2] == 1

        fracPickup_num = 1

        h_list = 1.
        w_list = np.arange(x_shape[3]) * 2. / (x_shape[3] - 1) - 1
        for i in range(fracPickup_num):
            idx = int(npr.rand() * len(w_list))
            if idx <= 0 or idx >= x_shape[3] - 1:
                continue
            beta = npr.rand() / 4.
            value0 = (beta * w_list[idx] + (1 - beta) * w_list[idx - 1])
            value1 = (beta * w_list[idx - 1] + (1 - beta) * w_list[idx])
            w_list[idx - 1] = value0
            w_list[idx] = value1

        grid = np.meshgrid(
            w_list,
            h_list,
            indexing='ij'
        )
        grid = np.stack(grid, axis=-1)
        grid = np.transpose(grid, (1, 0, 2))
        grid = np.expand_dims(grid, 0)
        grid = np.tile(grid, [x_shape[0], 1, 1, 1])
        grid = torch.from_numpy(grid).type(x.data.type())
        if self.cuda:
            grid = grid.cuda()
        self.grid = Variable(grid, requires_grad=False)

        x_offset = nn.functional.grid_sample(x, self.grid)

        return x_offset

class strLabelConverterForAttention(object):
    """Convert between str and label.

    NOTE:
        Insert `EOS` to the alphabet for attention.

    Args:
        alphabet (str): set of the possible characters.
        ignore_case (bool, default=True): whether or not to ignore all of the case.
    """

    def __init__(self, alphabet, sep):
        self._scanned_list = False
        self._out_of_list = ''
        self._ignore_case = True
        self.sep = sep
        self.alphabet = alphabet.split(sep)

        self.dict = {}
        for i, item in enumerate(self.alphabet):
            self.dict[item] = i

    def scan(self, text):
        # print(text)
        text_tmp = text
        text = []
        for i in range(len(text_tmp)):
            text_result = ''
            for j in range(len(text_tmp[i])):
                chara = text_tmp[i][j].lower() if self._ignore_case else text_tmp[i][j]
                if chara not in self.alphabet:
                    if chara in self._out_of_list:
                        continue
                    else:
                        self._out_of_list += chara
                        file_out_of_list = open("out_of_list.txt", "a+")
                        file_out_of_list.write(chara + "\n")
                        file_out_of_list.close()
                        print('" %s " is not in alphabet...' % chara)
                        continue
                else:
                    text_result += chara
            text.append(text_result)
        text_result = tuple(text)
        self._scanned_list = True
        return text_result

    def encode(self, text, scanned=True):
        """Support batch or single str.

        Args:
            text (str or list of str): texts to convert.

        Returns:
            torch.IntTensor [length_0 + length_1 + ... length_{n - 1}]: encoded texts.
            torch.IntTensor [n]: length of each text.
        """
        self._scanned_list = scanned
        if not self._scanned_list:
            text = self.scan(text)

        if isinstance(text, str):
            text = [
                self.dict[char.lower() if self._ignore_case else char]
                for char in text
            ]
            length = [len(text)]
        elif isinstance(text, collections.Iterable):
            length = [len(s) for s in text]
            text = ''.join(text)
            text, _ = self.encode(text)
        return (torch.LongTensor(text), torch.LongTensor(length))

    def decode(self, t, length):
        """Decode encoded texts back into strs.

        Args:
            torch.IntTensor [length_0 + length_1 + ... length_{n - 1}]: encoded texts.
            torch.IntTensor [n]: length of each text.

        Raises:
            AssertionError: when the texts and its length does not match.

        Returns:
            text (str or list of str): texts to convert.
        """
        if length.numel() == 1:
            length = length[0]
            assert t.numel() == length, "text with length: {} does not match declared length: {}".format(t.numel(), length)
            return ''.join([self.alphabet[i] for i in t])
        else:
            # batch mode
            assert t.numel() == length.sum(), "texts with length: {} does not match declared length: {}".format(t.numel(), length.sum())
            texts = []
            index = 0
            for i in range(length.numel()):
                l = length[i]
                texts.append(
                    self.decode(
                        t[index:index + l], torch.LongTensor([l])))
                index += l
            return texts

class averager(object):
    """Compute average for `torch.Variable` and `torch.Tensor`. """

    def __init__(self):
        self.reset()

    def add(self, v):
        if isinstance(v, Variable):
            count = v.data.numel()
            v = v.data.sum()
        elif isinstance(v, torch.Tensor):
            count = v.numel()
            v = v.sum()

        self.n_count += count
        self.sum += v

    def reset(self):
        self.n_count = 0
        self.sum = 0

    def val(self):
        res = 0
        if self.n_count != 0:
            res = self.sum / float(self.n_count)
        return res

def loadData(v, data):
    major, _ = get_torch_version()

    if major >= 1:
        v.resize_(data.size()).copy_(data)
    else:
        v.data.resize_(data.size()).copy_(data)

def get_torch_version():
    """
    Find pytorch version and return it as integers
    for major and minor versions
    """
    torch_version = str(torch.__version__).split(".")
    return int(torch_version[0]), int(torch_version[1])


class lmdbDataset(Dataset):

    def __init__(self, root=None, transform=None, reverse=False, alphabet='0123456789abcdefghijklmnopqrstuvwxyz'):
        self.env = lmdb.open(
            root,
            max_readers=1,
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False)

        if not self.env:
            print('cannot creat lmdb from %s' % (root))
            sys.exit(0)

        with self.env.begin(write=False) as txn:
            nSamples = int(txn.get('num-samples'.encode()))
            self.nSamples = nSamples

        self.transform = transform
        self.alphabet = alphabet
        self.reverse = reverse

    def __len__(self):
        return self.nSamples

    def __getitem__(self, index):
        assert index <= len(self), 'index range error'
        index += 1
        with self.env.begin(write=False) as txn:
            img_key = 'image-%09d' % index
            imgbuf = txn.get(img_key.encode())

            buf = six.BytesIO()
            buf.write(imgbuf)
            buf.seek(0)
            try:
                img = Image.open(buf).convert('L')
            except IOError:
                print('Corrupted image for %d' % index)
                return self[index + 1]

            label_key = 'label-%09d' % index
            label = str(txn.get(label_key.encode()).decode('utf-8'))

            label = ''.join(label[i] if label[i].lower() in self.alphabet else ''
                            for i in range(len(label)))
            if len(label) <= 0:
                return self[index + 1]
            if self.reverse:
                label_rev = label[-1::-1]
                label_rev += '$'
            label += '$'

            if self.transform is not None:
                img = self.transform(img)

        if self.reverse:
            return (img, label, label_rev)
        else:
            return (img, label)


class resizeNormalize(object):

    def __init__(self, size, interpolation=Image.BILINEAR):
        self.size = size
        self.interpolation = interpolation
        self.toTensor = transforms.ToTensor()

    def __call__(self, img):
        img = img.resize(self.size, self.interpolation)
        img = self.toTensor(img)
        img.sub_(0.5).div_(0.5)
        return img


class randomSequentialSampler(sampler.Sampler):

    def __init__(self, data_source, batch_size):
        self.num_samples = len(data_source)
        self.batch_size = batch_size

    def __len__(self):
        return self.num_samples

    def __iter__(self):
        n_batch = len(self) // self.batch_size
        tail = len(self) % self.batch_size
        index = torch.LongTensor(len(self)).fill_(0)
        for i in range(n_batch):
            random_start = random.randint(0, len(self) - self.batch_size)
            batch_index = random_start + torch.arange(0, self.batch_size)
            index[i * self.batch_size:(i + 1) * self.batch_size] = batch_index
        # deal with tail
        if tail:
            random_start = random.randint(0, len(self) - self.batch_size)
            tail_index = random_start + torch.arange(0, tail)
            index[(i + 1) * self.batch_size:] = tail_index

        return iter(index)








#!/usr/bin/env python

from __future__ import print_function

import random
import time

from progress.bar import (Bar, ChargingBar, FillingSquaresBar,
                          FillingCirclesBar, IncrementalBar, PixelBar,
                          ShadyBar)
from progress.spinner import (Spinner, PieSpinner, MoonSpinner, LineSpinner,
                              PixelSpinner)
from progress.counter import Counter, Countdown, Stack, Pie


def sleep():
    t = 0.01
    t += t * random.uniform(-0.1, 0.1)  # Add some variance
    time.sleep(t)


for bar_cls in (Bar, ChargingBar, FillingSquaresBar, FillingCirclesBar):
    suffix = '%(index)d/%(max)d [%(elapsed)d / %(eta)d / %(eta_td)s]'
    bar = bar_cls(bar_cls.__name__, suffix=suffix)
    for i in bar.iter(range(200)):
        sleep()

for bar_cls in (IncrementalBar, PixelBar, ShadyBar):
    suffix = '%(percent)d%% [%(elapsed_td)s / %(eta)d / %(eta_td)s]'
    bar = bar_cls(bar_cls.__name__, suffix=suffix)
    for i in bar.iter(range(200)):
        sleep()

for spin in (Spinner, PieSpinner, MoonSpinner, LineSpinner, PixelSpinner):
    for i in spin(spin.__name__ + ' ').iter(range(100)):
        sleep()
    print()

for singleton in (Counter, Countdown, Stack, Pie):
    for i in singleton(singleton.__name__ + ' ').iter(range(100)):
        sleep()
    print()

bar = IncrementalBar('Random', suffix='%(index)d')
for i in range(100):
    bar.goto(random.randint(0, 100))
    sleep()
bar.finish()



from libs.basic.basic_sequence import SeqAccumulator
import os
import cv2
import numpy as np
from libs.import_basic_utils import *
import json

class SuctionCifar10DMaker(SeqAccumulator):

    def process_this(self):
        self.mask_path = self.rgbd_path[0].replace('.png', '_mask.png')
        if not os.path.exists(self.mask_path):
            print('>>> %s does not exist >>> Skip' %self.mask_path)
            return False
        return super().process_this()
    def init_acc(self):
        super().init_acc()
        list_dir = os.path.join(self.args.root_dir, 'ade')
        if not os.path.exists(list_dir): os.makedirs(list_dir)
        self.train_list_file = open(os.path.join(list_dir,'training.odgt'), 'w+')
        self.test_list_file = open(os.path.join(list_dir, 'validation.odgt'), 'w+')

    def acc_sum(self):
        super().acc_sum()
        self.train_list_file.close()
        self.test_list_file.close()

    def process_single(self):
        # read mask
        mask = cv2.imread(self.mask_path, cv2.IMREAD_GRAYSCALE)
        mask_crop = ArrayUtils().rect_crop(mask, self.args.workspace.rect)

        rgbd_crop = self.rgbd.crop()

        suc_mask_org = (mask_crop > 150).astype('uint8')
        ker = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, self.args.suction_size)
        suc_mask = cv2.erode(suc_mask_org,ker)

        # ker = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (4*self.args.suction_size[0], 4*self.args.suction_size[1]))
        bound_mask = cv2.dilate(suc_mask_org, ker) - suc_mask

        # bg_mask = 1 - (bound_mask + suc_mask)

        mask_cmb = suc_mask + 2*bound_mask

        if self.args.show_steps:
            # cv2.imshow('mask_crop', mask_crop)
            # cv2.imshow('suc_mask', 255*suc_mask)
            # cv2.imshow('bg_mask', 255*bg_mask)
            # cv2.imshow('bound_mask', 255*bound_mask)
            cv2.imshow('mask_cmb', 50*mask_cmb)
            rgbd_crop.show(args=self.args, mask=bound_mask)
            if cv2.waitKey()==27: exit()


        if self.args.depth2norm:
            depth = ArrayUtils().get_mat_normal_map_U8(rgbd_crop.depth)
        else: depth = rgbd_crop.depth_U8()

        for j,color_order in enumerate(self.args.aug_color_orders):
            rgb = rgbd_crop.rgb[:,:,color_order]
            data = np.concatenate((rgb, depth), axis=2)
            height, width = rgb.shape[:2]


            # save images
            ann_filename = '%s_%d.png'%(self.filename, j)


            # update
            if self.to_train(self.acc['total']):
                num_train = self.acc['num_train']
                self.acc['mean'] = num_train / (num_train + 1) * self.acc['mean'] + \
                                   1 / (num_train + 1) * np.mean(data.astype('float') / 255, axis=(0, 1))
                self.acc['std'] = num_train / (num_train + 1) * self.acc['std'] + \
                                  1 / (num_train + 1) * np.std(data.astype('float') / 255, axis=(0, 1))
                self.acc['num_train'] += 1

                train_info = dict()
                fpath_img = os.path.join(self.args.root_dir, self.args.ann_rgb, 'training',ann_filename)
                train_info.update({'fpath_img':fpath_img})
                fpath_segm=os.path.join(self.args.root_dir, self.args.ann_mask, 'training', ann_filename)
                train_info.update({'fpath_segm': fpath_segm})
                fpath_depth = os.path.join(self.args.root_dir, self.args.ann_depth, 'training', ann_filename)
                # train_info.update({'fpath_depth': fpath_depth})
                train_info.update({'width': width, 'height':height})

                self.train_list_file.write(json.dumps(train_info) + '\n')

            else:
                self.acc['num_test'] += 1

                test_info = dict()
                fpath_img = os.path.join(self.args.root_dir, self.args.ann_rgb, 'validation', ann_filename)
                test_info.update({'fpath_img': fpath_img})
                fpath_segm = os.path.join(self.args.root_dir, self.args.ann_mask, 'validation', ann_filename)
                test_info.update({'fpath_segm': fpath_segm})
                fpath_depth = os.path.join(self.args.root_dir, self.args.ann_depth, 'validation', ann_filename)
                # test_info.update({'fpath_depth': fpath_depth})
                test_info.update({'width': width, 'height': height})

                self.test_list_file.write(json.dumps(test_info) + '\n')

            ArrayUtils().save_array_v3(array=rgb[:, :, ::-1], filepath=fpath_img)
            ArrayUtils().save_array_v3(array=depth, filepath=fpath_depth)
            ArrayUtils().save_array_v3(array=mask_cmb, filepath=fpath_segm)

            self.acc['total'] = self.acc['num_train'] + self.acc['num_test']



    def to_train(self, count):
        dur = self.args.train_val_div[0] + self.args.train_val_div[1]
        return (count % dur) < self.args.train_val_div[0]














if __name__ =='__main__':
    cfg_path = 'configs/segmentation/suction.cfg'

    SuctionCifar10DMaker(cfg_path=cfg_path).run()
    # PolygonShower(cfg_path=cfg_path).run()
    # OnMouseDrawGrip(cfg_path=cfg_path).run()
    # PolygonAnalyzer(cfg_path=cfg_path).run()

    # CombineCifar10Data(cfg_path=cfg_path).run()
    # CifarDepth2Norm(cfg_path=cfg_path).run()







import torch.nn as nn
import torch.nn.functional as F


class VGG_FeatureExtractor(nn.Module):
    """ FeatureExtractor of CRNN (https://arxiv.org/pdf/1507.05717.pdf) """

    def __init__(self, input_channel, output_channel=512):
        super(VGG_FeatureExtractor, self).__init__()
        self.output_channel = [int(output_channel / 8), int(output_channel / 4),
                               int(output_channel / 2), output_channel]  # [64, 128, 256, 512]
        self.ConvNet = nn.Sequential(
            nn.Conv2d(input_channel, self.output_channel[0], 3, 1, 1), nn.ReLU(True),
            nn.MaxPool2d(2, 2),  # 64x16x50
            nn.Conv2d(self.output_channel[0], self.output_channel[1], 3, 1, 1), nn.ReLU(True),
            nn.MaxPool2d(2, 2),  # 128x8x25
            nn.Conv2d(self.output_channel[1], self.output_channel[2], 3, 1, 1), nn.ReLU(True),  # 256x8x25
            nn.Conv2d(self.output_channel[2], self.output_channel[2], 3, 1, 1), nn.ReLU(True),
            nn.MaxPool2d((2, 1), (2, 1)),  # 256x4x25
            nn.Conv2d(self.output_channel[2], self.output_channel[3], 3, 1, 1, bias=False),
            nn.BatchNorm2d(self.output_channel[3]), nn.ReLU(True),  # 512x4x25
            nn.Conv2d(self.output_channel[3], self.output_channel[3], 3, 1, 1, bias=False),
            nn.BatchNorm2d(self.output_channel[3]), nn.ReLU(True),
            nn.MaxPool2d((2, 1), (2, 1)),  # 512x2x25
            nn.Conv2d(self.output_channel[3], self.output_channel[3], 2, 1, 0), nn.ReLU(True))  # 512x1x24

    def forward(self, input):
        return self.ConvNet(input)


class RCNN_FeatureExtractor(nn.Module):
    """ FeatureExtractor of GRCNN (https://papers.nips.cc/paper/6637-gated-recurrent-convolution-neural-network-for-ocr.pdf) """

    def __init__(self, input_channel, output_channel=512):
        super(RCNN_FeatureExtractor, self).__init__()
        self.output_channel = [int(output_channel / 8), int(output_channel / 4),
                               int(output_channel / 2), output_channel]  # [64, 128, 256, 512]
        self.ConvNet = nn.Sequential(
            nn.Conv2d(input_channel, self.output_channel[0], 3, 1, 1), nn.ReLU(True),
            nn.MaxPool2d(2, 2),  # 64 x 16 x 50
            GRCL(self.output_channel[0], self.output_channel[0], num_iteration=5, kernel_size=3, pad=1),
            nn.MaxPool2d(2, 2),  # 64 x 8 x 25
            GRCL(self.output_channel[0], self.output_channel[1], num_iteration=5, kernel_size=3, pad=1),
            nn.MaxPool2d(2, (2, 1), (0, 1)),  # 128 x 4 x 26
            GRCL(self.output_channel[1], self.output_channel[2], num_iteration=5, kernel_size=3, pad=1),
            nn.MaxPool2d(2, (2, 1), (0, 1)),  # 256 x 2 x 27
            nn.Conv2d(self.output_channel[2], self.output_channel[3], 2, 1, 0, bias=False),
            nn.BatchNorm2d(self.output_channel[3]), nn.ReLU(True))  # 512 x 1 x 26

    def forward(self, input):
        return self.ConvNet(input)


class ResNet_FeatureExtractor(nn.Module):
    """ FeatureExtractor of FAN (http://openaccess.thecvf.com/content_ICCV_2017/papers/Cheng_Focusing_Attention_Towards_ICCV_2017_paper.pdf) """

    def __init__(self, input_channel, output_channel=512):
        super(ResNet_FeatureExtractor, self).__init__()
        self.ConvNet = ResNet(input_channel, output_channel, BasicBlock, [1, 2, 5, 3])

    def forward(self, input):
        return self.ConvNet(input)


# For Gated RCNN
class GRCL(nn.Module):

    def __init__(self, input_channel, output_channel, num_iteration, kernel_size, pad):
        super(GRCL, self).__init__()
        self.wgf_u = nn.Conv2d(input_channel, output_channel, 1, 1, 0, bias=False)
        self.wgr_x = nn.Conv2d(output_channel, output_channel, 1, 1, 0, bias=False)
        self.wf_u = nn.Conv2d(input_channel, output_channel, kernel_size, 1, pad, bias=False)
        self.wr_x = nn.Conv2d(output_channel, output_channel, kernel_size, 1, pad, bias=False)

        self.BN_x_init = nn.BatchNorm2d(output_channel)

        self.num_iteration = num_iteration
        self.GRCL = [GRCL_unit(output_channel) for _ in range(num_iteration)]
        self.GRCL = nn.Sequential(*self.GRCL)

    def forward(self, input):
        """ The input of GRCL is consistant over time t, which is denoted by u(0)
        thus wgf_u / wf_u is also consistant over time t.
        """
        wgf_u = self.wgf_u(input)
        wf_u = self.wf_u(input)
        x = F.relu(self.BN_x_init(wf_u))

        for i in range(self.num_iteration):
            x = self.GRCL[i](wgf_u, self.wgr_x(x), wf_u, self.wr_x(x))

        return x


class GRCL_unit(nn.Module):

    def __init__(self, output_channel):
        super(GRCL_unit, self).__init__()
        self.BN_gfu = nn.BatchNorm2d(output_channel)
        self.BN_grx = nn.BatchNorm2d(output_channel)
        self.BN_fu = nn.BatchNorm2d(output_channel)
        self.BN_rx = nn.BatchNorm2d(output_channel)
        self.BN_Gx = nn.BatchNorm2d(output_channel)

    def forward(self, wgf_u, wgr_x, wf_u, wr_x):
        G_first_term = self.BN_gfu(wgf_u)
        G_second_term = self.BN_grx(wgr_x)
        G = F.sigmoid(G_first_term + G_second_term)

        x_first_term = self.BN_fu(wf_u)
        x_second_term = self.BN_Gx(self.BN_rx(wr_x) * G)
        x = F.relu(x_first_term + x_second_term)

        return x


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = self._conv3x3(inplanes, planes)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = self._conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def _conv3x3(self, in_planes, out_planes, stride=1):
        "3x3 convolution with padding"
        return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                         padding=1, bias=False)

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, input_channel, output_channel, block, layers):
        super(ResNet, self).__init__()

        self.output_channel_block = [int(output_channel / 4), int(output_channel / 2), output_channel, output_channel]

        self.inplanes = int(output_channel / 8)
        self.conv0_1 = nn.Conv2d(input_channel, int(output_channel / 16),
                                 kernel_size=3, stride=1, padding=1, bias=False)
        self.bn0_1 = nn.BatchNorm2d(int(output_channel / 16))
        self.conv0_2 = nn.Conv2d(int(output_channel / 16), self.inplanes,
                                 kernel_size=3, stride=1, padding=1, bias=False)
        self.bn0_2 = nn.BatchNorm2d(self.inplanes)
        self.relu = nn.ReLU(inplace=True)

        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2, padding=0)
        self.layer1 = self._make_layer(block, self.output_channel_block[0], layers[0])
        self.conv1 = nn.Conv2d(self.output_channel_block[0], self.output_channel_block[
                               0], kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(self.output_channel_block[0])

        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2, padding=0)
        self.layer2 = self._make_layer(block, self.output_channel_block[1], layers[1], stride=1)
        self.conv2 = nn.Conv2d(self.output_channel_block[1], self.output_channel_block[
                               1], kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(self.output_channel_block[1])

        self.maxpool3 = nn.MaxPool2d(kernel_size=2, stride=(2, 1), padding=(0, 1))
        self.layer3 = self._make_layer(block, self.output_channel_block[2], layers[2], stride=1)
        self.conv3 = nn.Conv2d(self.output_channel_block[2], self.output_channel_block[
                               2], kernel_size=3, stride=1, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.output_channel_block[2])

        self.layer4 = self._make_layer(block, self.output_channel_block[3], layers[3], stride=1)
        self.conv4_1 = nn.Conv2d(self.output_channel_block[3], self.output_channel_block[
                                 3], kernel_size=2, stride=(2, 1), padding=(0, 1), bias=False)
        self.bn4_1 = nn.BatchNorm2d(self.output_channel_block[3])
        self.conv4_2 = nn.Conv2d(self.output_channel_block[3], self.output_channel_block[
                                 3], kernel_size=2, stride=1, padding=0, bias=False)
        self.bn4_2 = nn.BatchNorm2d(self.output_channel_block[3])

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv0_1(x)
        x = self.bn0_1(x)
        x = self.relu(x)
        x = self.conv0_2(x)
        x = self.bn0_2(x)
        x = self.relu(x)

        x = self.maxpool1(x)
        x = self.layer1(x)
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.maxpool2(x)
        x = self.layer2(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)

        x = self.maxpool3(x)
        x = self.layer3(x)
        x = self.conv3(x)
        x = self.bn3(x)
        x = self.relu(x)

        x = self.layer4(x)
        x = self.conv4_1(x)
        x = self.bn4_1(x)
        x = self.relu(x)
        x = self.conv4_2(x)
        x = self.bn4_2(x)
        x = self.relu(x)

        return x



from ..import_basic_utils import *
# from ..utils.makedb_utilts import *
from ..basic.basic_onmouse import BasSeqImOnMouseLabel
from ..basic.basic_objects import BasObj
from ..basic.basic_objects import BasSeqImObj, BasPoly, BasJson
from ..basic.basic_sequence import SeqAccumulator, BasSeqImCallBack
import os, cv2


def show_polygon(rgbd, json_path, args):
    if json_path is None: return rgbd.crop_rgb()
    return visualize_polygon(im=rgbd.crop_rgb(), json_path=json_path,
                             text_scale=args.text_scale, text_thick=args.text_thick,
                             rect=args.text_rect, space=args.text_space, alpha=args.text_alpha, up2down=args.up2down)

def polys2masks(polys, im_size):
    mask = np.zeros(im_size, np.uint16)
    num_poly = len(polys)
    for j in range(num_poly):
        poly = np.array(polys[j]).astype('int32')

        cv2.fillPoly(mask,poly.reshape((1, )+ poly.shape),j+1000)
    return mask

class DrawPolyon(BasObj):
    def get_ann(self, inputs, filename='unname', title='viewer'):
        json_path = os.path.join(self.args.root_dir, self.args.ann_dir, filename + '.json')
        if not os.path.exists(json_path): return None
        return json_path

class OnMousePolygonDrawer(BasSeqImOnMouseLabel):
    def start_on_mouse(self, filename, **kwargs):
        ann_dict = super().start_on_mouse(filename=filename)
        self.disp_colors = []
        if ann_dict is None: self.class_list, self.polygons = [], []
        else:
            self.class_list, self.polygons = ann_dict['Classes'], ann_dict['Polygons']
            for cls in self.class_list:
                for key in self.key_dict:
                    if self.key_dict[key]['cls'] != cls: continue
                    self.disp_colors.append(self.key_dict[key]['color'])


    def show_draws(self,rgbd):
        self.im = PolyUtils().draw_poly(self.im, self.click_locs, thick=self.args.line_thick,
                        color=self.disp_colors[-1])  # permenantly show


    def process_click_locs(self, rgbd):
        self.polygons.append(self.click_locs)

        # region class
        promtstr = 'Class? \n'
        for key in self.key_dict:
            promtstr += '%s: \t%s\n' % (key, self.key_dict[key]['cls'])
        print(promtstr)

        while True:
            key = chr(cv2.waitKey())
            if key in self.key_dict:
                break
            else:
                print('wrong key, press other key ... ')
        self.class_list.append(self.key_dict[key]['cls'])
        self.disp_colors.append(self.key_dict[key]['color'])
        ProcUtils().clscr()

        # grip = GRIP(pts = self.click_locs[:2])
        # self.im = rgbd.disp(args=self.args, grips=GRIPS(grips=grip))

        self.show_draws(rgbd)

    def stop_on_mouse(self, rgbd, filename):
        super().stop_on_mouse(rgbd=rgbd, filename=filename)
        json_name, _ = os.path.splitext(filename)
        json_name += '.json'
        JsonUtils().save_poly_json(os.path.join(self.args.root_dir, self.args.ann_poly,json_name), self.im_info, self.class_list, self.polygons, )


class PolygonShower(BasSeqImCallBack, BasPoly):
    def process_single(self):
        super().process_single()
        self.show_polys(rgbd=self.rgbd, filename=self.filename)

    def show_polys(self, rgbd, filename='unnamed'):
        ann_polygon = os.path.join(self.args.root_dir, self.args.ann_poly)
        json_path = os.path.join(ann_polygon, filename + '.json')
        if not os.path.exists(json_path):
            print('%s does not exist ...' %json_path)
            rgbd.show(args=self.args)
        else:
            out = self.draw_polys_from_json(json_path, rgbd.bgr())
            cv2.imshow('viewer', out)
            # return self.draw_polys_on_rgbd(json_path, rgbd)


class PolygonAnalyzer(SeqAccumulator, BasJson):
    def process_single(self):
        super().process_single()
        in_dir, self.filename = os.path.split(self.rgbd_path[0])
        self.get_rgbd_from_path(self.rgbd_path)
        # ----------------------------->>>>>>>>>> Main
        ann_polygon = os.path.join(self.args.root_dir, self.args.ann_polygon)
        json_path = os.path.join(ann_polygon, self.filename).replace('.png', '.json')
        if not os.path.exists(json_path):
            print('%s does not exist ...' % json_path)
            return

        ann_dict = self.read_json(json_path=json_path)
        self.analyze_classes(ann_dict['Classes'])


    def analyze_classes(self, classes):
        anly = dict()
        for cls in classes:
            if cls not in anly: anly.update({cls: 1})
            else : anly[cls] += 1

        for cls in anly:
            print('cls %s: %d' %(cls, anly[cls]))







from __future__ import division
""" 
Creates a ResNeXt Model as defined in:
Xie, S., Girshick, R., Dollar, P., Tu, Z., & He, K. (2016). 
Aggregated residual transformations for deep neural networks. 
arXiv preprint arXiv:1611.05431.
import from https://github.com/prlz77/ResNeXt.pytorch/blob/master/models/model.py
"""
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init

__all__ = ['resnext']

class ResNeXtBottleneck(nn.Module):
    """
    RexNeXt bottleneck type C (https://github.com/facebookresearch/ResNeXt/blob/master/models/resnext.lua)
    """
    def __init__(self, in_channels, out_channels, stride, cardinality, widen_factor):
        """ Constructor
        Args:
            in_channels: input channel dimensionality
            out_channels: output channel dimensionality
            stride: conv stride. Replaces pooling layer.
            cardinality: num of convolution groups.
            widen_factor: factor to reduce the input dimensionality before convolution.
        """
        super(ResNeXtBottleneck, self).__init__()
        D = cardinality * out_channels // widen_factor
        self.conv_reduce = nn.Conv2d(in_channels, D, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn_reduce = nn.BatchNorm2d(D)
        self.conv_conv = nn.Conv2d(D, D, kernel_size=3, stride=stride, padding=1, groups=cardinality, bias=False)
        self.bn = nn.BatchNorm2d(D)
        self.conv_expand = nn.Conv2d(D, out_channels, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn_expand = nn.BatchNorm2d(out_channels)

        self.shortcut = nn.Sequential()
        if in_channels != out_channels:
            self.shortcut.add_module('shortcut_conv', nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, padding=0, bias=False))
            self.shortcut.add_module('shortcut_bn', nn.BatchNorm2d(out_channels))

    def forward(self, x):
        bottleneck = self.conv_reduce.forward(x)
        bottleneck = F.relu(self.bn_reduce.forward(bottleneck), inplace=True)
        bottleneck = self.conv_conv.forward(bottleneck)
        bottleneck = F.relu(self.bn.forward(bottleneck), inplace=True)
        bottleneck = self.conv_expand.forward(bottleneck)
        bottleneck = self.bn_expand.forward(bottleneck)
        residual = self.shortcut.forward(x)
        return F.relu(residual + bottleneck, inplace=True)


class CifarResNeXt(nn.Module):
    """
    ResNext optimized for the Cifar dataset, as specified in
    https://arxiv.org/pdf/1611.05431.pdf
    """
    def __init__(self, cardinality, depth, num_classes, widen_factor=4, dropRate=0):
        """ Constructor
        Args:
            cardinality: number of convolution groups.
            depth: number of layers.
            num_classes: number of classes
            widen_factor: factor to adjust the channel dimensionality
        """
        super(CifarResNeXt, self).__init__()
        self.cardinality = cardinality
        self.depth = depth
        self.block_depth = (self.depth - 2) // 9
        self.widen_factor = widen_factor
        self.num_classes = num_classes
        self.output_size = 64
        self.stages = [64, 64 * self.widen_factor, 128 * self.widen_factor, 256 * self.widen_factor]

        self.conv_1_3x3 = nn.Conv2d(3, 64, 3, 1, 1, bias=False)
        self.bn_1 = nn.BatchNorm2d(64)
        self.stage_1 = self.block('stage_1', self.stages[0], self.stages[1], 1)
        self.stage_2 = self.block('stage_2', self.stages[1], self.stages[2], 2)
        self.stage_3 = self.block('stage_3', self.stages[2], self.stages[3], 2)
        self.classifier = nn.Linear(1024, num_classes)
        init.kaiming_normal(self.classifier.weight)

        for key in self.state_dict():
            if key.split('.')[-1] == 'weight':
                if 'conv' in key:
                    init.kaiming_normal(self.state_dict()[key], mode='fan_out')
                if 'bn' in key:
                    self.state_dict()[key][...] = 1
            elif key.split('.')[-1] == 'bias':
                self.state_dict()[key][...] = 0

    def block(self, name, in_channels, out_channels, pool_stride=2):
        """ Stack n bottleneck modules where n is inferred from the depth of the network.
        Args:
            name: string name of the current block.
            in_channels: number of input channels
            out_channels: number of output channels
            pool_stride: factor to reduce the spatial dimensionality in the first bottleneck of the block.
        Returns: a Module consisting of n sequential bottlenecks.
        """
        block = nn.Sequential()
        for bottleneck in range(self.block_depth):
            name_ = '%s_bottleneck_%d' % (name, bottleneck)
            if bottleneck == 0:
                block.add_module(name_, ResNeXtBottleneck(in_channels, out_channels, pool_stride, self.cardinality,
                                                          self.widen_factor))
            else:
                block.add_module(name_,
                                 ResNeXtBottleneck(out_channels, out_channels, 1, self.cardinality, self.widen_factor))
        return block

    def forward(self, x):
        x = self.conv_1_3x3.forward(x)
        x = F.relu(self.bn_1.forward(x), inplace=True)
        x = self.stage_1.forward(x)
        x = self.stage_2.forward(x)
        x = self.stage_3.forward(x)
        x = F.avg_pool2d(x, 8, 1)
        x = x.view(-1, 1024)
        return self.classifier(x)

def resnext(**kwargs):
    """Constructs a ResNeXt.
    """
    model = CifarResNeXt(**kwargs)
    return model


from libs.import_basic_utils import *
from libs.basic.basic_objects import BasDetectObj
from detectron2.config import get_cfg
from detectron2 import model_zoo
from detectron2.engine import DefaultPredictor
import cv2
from detectron2.utils.visualizer import Visualizer
from detectron2.data import DatasetCatalog, MetadataCatalog
import numpy as np

class MaskRCNNDetectron(BasDetectObj):
    def get_model(self):
        cfg = get_cfg()
        cfg.merge_from_file(model_zoo.get_config_file(self.args.merge_from_file))
        cfg.DATASETS.TEST = self.args.test_dataset
        cfg.DATALOADER.NUM_WORKERS = self.args.num_workers
        cfg.MODEL.ROI_HEADS.BATCH_SIZE_PER_IMAGE = self.args.batch_size_per_image
        cfg.MODEL.ROI_HEADS.NUM_CLASSES = self.args.num_classes  # only has one class (ballon)
        cfg.MODEL.ANCHOR_GENERATOR.SIZES = self.args.anchor_sizes
        cfg.MODEL.WEIGHTS = self.args.checkpoint
        cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = self.args.score_thresh_test

        self.detector = DefaultPredictor(cfg)

    def predict_array(self, anArray, workspace=None):
        height, width = anArray.shape[:2]
        if workspace is None:
            ret = self.detector(anArray)["instances"].to("cpu")
            boxes = ret.pred_boxes.tensor.numpy()
            masks = ret.pred_masks.numpy().astype('uint8')
            classes = ret.pred_classes.numpy()
            scores = ret.scores.numpy()

            return {'boxes': boxes, 'masks': masks, 'classes':classes, 'scores': scores}

        if workspace is not None:
            array_crop = anArray[workspace.rect.top: workspace.rect.bottom, workspace.rect.left:workspace.rect.right, :]
            ret = self.predict_array(array_crop)

            num_detected= len(ret)
            if num_detected==0: return  ret

            ret['boxes'][0] +=  workspace.rect.left
            ret['boxes'][2] += workspace.rect.left
            ret['boxes'][1] += workspace.rect.top
            ret['boxes'][3] += workspace.rect.top

            right = width-workspace.rect.right
            bottom = height-workspace.rect.bottom

            new_masks = []
            for i in range(num_detected):
                mask = ret['masks'][i, :,:]
                mask = cv2.copyMakeBorder(mask,workspace.rect.top, bottom,workspace.rect.left, right, cv2.BORDER_CONSTANT, value=0)
                new_masks.append(mask)

            ret['mask'] = np.vstack(new_masks)

            return ret






    def show_detect_results(self, im, detect_rets):
        boxes = detect_rets['boxes']
        masks = detect_rets['masks']
        classes = detect_rets['classes']
        scores = detect_rets['scores']
        out = np.copy(im)
        obj_ind = 0
        num_detected = boxes.shape[0]
        if num_detected==0: return im

        for i in range(num_detected):
            cls = classes[i]
            score = scores[i]
            if cls == '__background__': continue
            box = boxes[i,:].astype('int')
            mask = masks[i, :,:]

            color = ProcUtils().get_color(obj_ind)

            left, top, right, bottom = box

            loc = np.where(mask)

            out[loc] = 0.8 * out[loc] + tuple(0.2 * np.array(color))
            cv2.rectangle(out, (left, top), (right, bottom), color, 2)
            cv2.putText(out, '%s: %d' %(cls, int(100*score)), (left, top-3), cv2.FONT_HERSHEY_SIMPLEX, 0.5, color)

            obj_ind += 1

        return out



    def visualize_ret(self, rgb, ret):


        for d in ["val"]:
            # DatasetCatalog.register("prop_" + d, lambda d=d: get_dicts("prop/" + d))
            MetadataCatalog.get("prop_" + d).set(thing_classes=["prop"])
        prop_metadata = MetadataCatalog.get("prop_train")

        v = Visualizer(rgb,
                       metadata=prop_metadata,
                       scale=0.8
                       # instance_mode=ColorMode.IMAGE_BW  # remove the colors of unsegmented pixels
                       )
        # print(outputs["instances"])
        v = v.draw_instance_predictions(ret)
        print(ret.pred_classes)
        print(ret.pred_boxes)
        return v.get_image()[:,:,::-1]



#!/usr/bin/env python

from setuptools import setup

import progress


setup(
    name='progress',
    version=progress.__version__,
    description='Easy to use progress bars',
    long_description=open('README.rst').read(),
    author='Giorgos Verigakis',
    author_email='verigak@gmail.com',
    url='http://github.com/verigak/progress/',
    license='ISC',
    packages=['progress'],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)



import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import torchvision
import torchvision.transforms as transforms
import numpy as np
from .misc import *   

__all__ = ['make_image', 'show_batch', 'show_mask', 'show_mask_single']

# functions to show an image
def make_image(img, mean=(0,0,0), std=(1,1,1)):
    for i in range(0, 3):
        img[i] = img[i] * std[i] + mean[i]    # unnormalize
    npimg = img.numpy()
    return np.transpose(npimg, (1, 2, 0))

def gauss(x,a,b,c):
    return torch.exp(-torch.pow(torch.add(x,-b),2).div(2*c*c)).mul(a)

def colorize(x):
    ''' Converts a one-channel grayscale image to a color heatmap image '''
    if x.dim() == 2:
        torch.unsqueeze(x, 0, out=x)
    if x.dim() == 3:
        cl = torch.zeros([3, x.size(1), x.size(2)])
        cl[0] = gauss(x,.5,.6,.2) + gauss(x,1,.8,.3)
        cl[1] = gauss(x,1,.5,.3)
        cl[2] = gauss(x,1,.2,.3)
        cl[cl.gt(1)] = 1
    elif x.dim() == 4:
        cl = torch.zeros([x.size(0), 3, x.size(2), x.size(3)])
        cl[:,0,:,:] = gauss(x,.5,.6,.2) + gauss(x,1,.8,.3)
        cl[:,1,:,:] = gauss(x,1,.5,.3)
        cl[:,2,:,:] = gauss(x,1,.2,.3)
    return cl

def show_batch(images, Mean=(2, 2, 2), Std=(0.5,0.5,0.5)):
    images = make_image(torchvision.utils.make_grid(images), Mean, Std)
    plt.imshow(images)
    plt.show()


def show_mask_single(images, mask, Mean=(2, 2, 2), Std=(0.5,0.5,0.5)):
    im_size = images.size(2)

    # save for adding mask
    im_data = images.clone()
    for i in range(0, 3):
        im_data[:,i,:,:] = im_data[:,i,:,:] * Std[i] + Mean[i]    # unnormalize

    images = make_image(torchvision.utils.make_grid(images), Mean, Std)
    plt.subplot(2, 1, 1)
    plt.imshow(images)
    plt.axis('off')

    # for b in range(mask.size(0)):
    #     mask[b] = (mask[b] - mask[b].min())/(mask[b].max() - mask[b].min())
    mask_size = mask.size(2)
    # print('Max %f Min %f' % (mask.max(), mask.min()))
    mask = (upsampling(mask, scale_factor=im_size/mask_size))
    # mask = colorize(upsampling(mask, scale_factor=im_size/mask_size))
    # for c in range(3):
    #     mask[:,c,:,:] = (mask[:,c,:,:] - Mean[c])/Std[c]

    # print(mask.size())
    mask = make_image(torchvision.utils.make_grid(0.3*im_data+0.7*mask.expand_as(im_data)))
    # mask = make_image(torchvision.utils.make_grid(0.3*im_data+0.7*mask), Mean, Std)
    plt.subplot(2, 1, 2)
    plt.imshow(mask)
    plt.axis('off')

def show_mask(images, masklist, Mean=(2, 2, 2), Std=(0.5,0.5,0.5)):
    im_size = images.size(2)

    # save for adding mask
    im_data = images.clone()
    for i in range(0, 3):
        im_data[:,i,:,:] = im_data[:,i,:,:] * Std[i] + Mean[i]    # unnormalize

    images = make_image(torchvision.utils.make_grid(images), Mean, Std)
    plt.subplot(1+len(masklist), 1, 1)
    plt.imshow(images)
    plt.axis('off')

    for i in range(len(masklist)):
        mask = masklist[i].data.cpu()
        # for b in range(mask.size(0)):
        #     mask[b] = (mask[b] - mask[b].min())/(mask[b].max() - mask[b].min())
        mask_size = mask.size(2)
        # print('Max %f Min %f' % (mask.max(), mask.min()))
        mask = (upsampling(mask, scale_factor=im_size/mask_size))
        # mask = colorize(upsampling(mask, scale_factor=im_size/mask_size))
        # for c in range(3):
        #     mask[:,c,:,:] = (mask[:,c,:,:] - Mean[c])/Std[c]

        # print(mask.size())
        mask = make_image(torchvision.utils.make_grid(0.3*im_data+0.7*mask.expand_as(im_data)))
        # mask = make_image(torchvision.utils.make_grid(0.3*im_data+0.7*mask), Mean, Std)
        plt.subplot(1+len(masklist), 1, i+2)
        plt.imshow(mask)
        plt.axis('off')



# x = torch.zeros(1, 3, 3)
# out = colorize(x)
# out_im = make_image(out)
# plt.imshow(out_im)
# plt.show()


import torch
from ..utils.deeptext_utils import CTCLabelConverter, AttnLabelConverter
from .deeptext_model.model import Model
from KAI.utils.deeptext_utils import RawDataset, AlignCollate
import torch.nn.functional as F
import os

class TextRecognizer():
    def __int__(self, args):

        self.load_params(args)
        self.load_model()

    def load_params(self, args):
        self.args = args
        if 'CTC' in self.args['prediction']:
            self.converter = CTCLabelConverter(self.args['character'])
        else:
            self.converter = AttnLabelConverter(self.args['character'])
        self.args.update({'num_class' : len(self.converter.character)})
        checpoint_file = '%s-%s-%s-%s.pth' \
                     %(self.args['transformation'], self.args['featureextraction'],
                       self.args['sequencemodeling'],self.args['prediction'])

        self.model_path = os.path.join(self.args['text_recogizer_checkpoint_dir'], checpoint_file)

    def load_model(self):
        model = Model(self.args)
        self.model = torch.nn.DataParallel(model).to('cuda')
        self.model.load_state_dict(torch.load(self.model_path, map_location='cuda'))


    def predict_arrays(self, im_list, labels='unnamed'):

        AlignCollate_demo = AlignCollate(imgH=self.args['imgh'], imgW=self.args['imgw'],
                                         keep_ratio_with_pad=self.args['pad'])
        demo_data = RawDataset(im_list, self.args, labels)
        demo_loader = torch.utils.data.DataLoader(
            demo_data, batch_size=self.args['batch_size'],
            shuffle=False,
            num_workers=self.args['workers'],
            collate_fn=AlignCollate_demo, pin_memory=True)

        # predict
        self.model.eval()
        text_outs = []
        score_outs = []
        # with torch.no_grad():
        for image_tensors, image_path_list in demo_loader:
            batch_size = image_tensors.size(0)
            image = image_tensors.to('cuda')
            # For max length prediction
            length_for_pred = torch.IntTensor([self.args['batch_max_length']] * batch_size).to('cuda')
            text_for_pred = torch.LongTensor(batch_size, self.args['batch_max_length'] + 1).fill_(0).to('cuda')

            if 'CTC' in self.args['prediction']:
                preds = self.model(image, text_for_pred).log_softmax(2)

                # Select max probabilty (greedy decoding) then decode index to character
                preds_size = torch.IntTensor([preds.size(1)] * batch_size)
                _, preds_index = preds.max(2)
                preds_index = preds_index.view(-1)
                preds_str = self.converter.decode(preds_index.data, preds_size.data)

            else:
                preds = self.model(image, text_for_pred, is_train=False)

                # select max probabilty (greedy decoding) then decode index to character
                _, preds_index = preds.max(2)
                preds_str = self.converter.decode(preds_index, length_for_pred)

            if self.args['show_steps']:
                print('-' * 80)
                print('image_path \tpredict_label \t\tconfidence score')
                print('-' * 80)

            preds_prob = F.softmax(preds, dim=2)
            preds_max_prob, _ = preds_prob.max(dim=2)
            angles = [90, -90]
            for img_name, pred, pred_max_prob in zip(image_path_list, preds_str, preds_max_prob):

                if 'Attn' in self.args['prediction']:
                    pred_EOS = pred.find('[s]')
                    pred = pred[:pred_EOS]  # prune after "end of sentence" token ([s])
                    pred_max_prob = pred_max_prob[:pred_EOS]

                confidence_score = pred_max_prob.cumprod(dim=0)[-1]
                if self.args['show_steps']:
                    print('%s \t\t%s \t\t\t%.4f' % (img_name, pred, confidence_score))

                text_outs.append(pred)
                score_outs.append(float(confidence_score))



        return text_outs, score_outs











import torch
import torch.nn as nn
import torch.nn.functional as F
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


class Attention(nn.Module):

    def __init__(self, input_size, hidden_size, num_classes):
        super(Attention, self).__init__()
        self.attention_cell = AttentionCell(input_size, hidden_size, num_classes)
        self.hidden_size = hidden_size
        self.num_classes = num_classes
        self.generator = nn.Linear(hidden_size, num_classes)

    def _char_to_onehot(self, input_char, onehot_dim=38):
        input_char = input_char.unsqueeze(1)
        batch_size = input_char.size(0)
        one_hot = torch.FloatTensor(batch_size, onehot_dim).zero_().to(device)
        one_hot = one_hot.scatter_(1, input_char, 1)
        return one_hot

    def forward(self, batch_H, text, is_train=True, batch_max_length=25):
        """
        input:
            batch_H : contextual_feature H = hidden state of encoder. [batch_size x num_steps x num_classes]
            text : the text-index of each image. [batch_size x (max_length+1)]. +1 for [GO] token. text[:, 0] = [GO].
        output: probability distribution at each step [batch_size x num_steps x num_classes]
        """
        batch_size = batch_H.size(0)
        num_steps = batch_max_length + 1  # +1 for [s] at end of sentence.

        output_hiddens = torch.FloatTensor(batch_size, num_steps, self.hidden_size).fill_(0).to(device)
        hidden = (torch.FloatTensor(batch_size, self.hidden_size).fill_(0).to(device),
                  torch.FloatTensor(batch_size, self.hidden_size).fill_(0).to(device))

        if is_train:
            for i in range(num_steps):
                # one-hot vectors for a i-th char. in a batch
                char_onehots = self._char_to_onehot(text[:, i], onehot_dim=self.num_classes)
                # hidden : decoder's hidden s_{t-1}, batch_H : encoder's hidden H, char_onehots : one-hot(y_{t-1})
                hidden, alpha = self.attention_cell(hidden, batch_H, char_onehots)
                output_hiddens[:, i, :] = hidden[0]  # LSTM hidden index (0: hidden, 1: Cell)
            probs = self.generator(output_hiddens)

        else:
            targets = torch.LongTensor(batch_size).fill_(0).to(device)  # [GO] token
            probs = torch.FloatTensor(batch_size, num_steps, self.num_classes).fill_(0).to(device)

            for i in range(num_steps):
                char_onehots = self._char_to_onehot(targets, onehot_dim=self.num_classes)
                hidden, alpha = self.attention_cell(hidden, batch_H, char_onehots)
                probs_step = self.generator(hidden[0])
                probs[:, i, :] = probs_step
                _, next_input = probs_step.max(1)
                targets = next_input

        return probs  # batch_size x num_steps x num_classes


class AttentionCell(nn.Module):

    def __init__(self, input_size, hidden_size, num_embeddings):
        super(AttentionCell, self).__init__()
        self.i2h = nn.Linear(input_size, hidden_size, bias=False)
        self.h2h = nn.Linear(hidden_size, hidden_size)  # either i2i or h2h should have bias
        self.score = nn.Linear(hidden_size, 1, bias=False)
        self.rnn = nn.LSTMCell(input_size + num_embeddings, hidden_size)
        self.hidden_size = hidden_size

    def forward(self, prev_hidden, batch_H, char_onehots):
        # [batch_size x num_encoder_step x num_channel] -> [batch_size x num_encoder_step x hidden_size]
        batch_H_proj = self.i2h(batch_H)
        prev_hidden_proj = self.h2h(prev_hidden[0]).unsqueeze(1)
        e = self.score(torch.tanh(batch_H_proj + prev_hidden_proj))  # batch_size x num_encoder_step * 1

        alpha = F.softmax(e, dim=1)
        context = torch.bmm(alpha.permute(0, 2, 1), batch_H).squeeze(1)  # batch_size x num_channel
        concat_context = torch.cat([context, char_onehots], 1)  # batch_size x (num_channel + num_embedding)
        cur_hidden = self.rnn(concat_context, prev_hidden)
        return cur_hidden, alpha



from KAI.utils.proc_utils import *
from KAI.utils.data_processing import *
from KAI.utils.makedb_utilts import *
from KAI.dataset.polygon import *
from glob import glob
from scipy.ndimage import rotate
from KAI.basic.basic_objects import BasObj
import json

class GripDBer(BasObj):

    def load_params(self, args):
        super().load_params(args=args)
        aug_bg_dir = os.path.join(args.root_dir, args.aug_bg_dir)
        assert os.path.exists(aug_bg_dir)
        all_bg_path = glob(os.path.join(aug_bg_dir, '*'))
        self.all_bg = [cv2.resize(cv2.imread(p), args.out_size[::-1]) for p in all_bg_path]
        self.all_angle = list(np.arange(0,360, args.step_angle))
        self.num_aug = min(len(self.all_bg), len(self.all_angle))

        # load
        self.info_dir  = os.path.join(args.root_dir, args.info_dir)
        self.count_file = os.path.join(self.info_dir, self.args.count_file)
        self.rgb_mean_file = os.path.join(self.info_dir, self.args.rgb_mean_file)
        self.std_mean_file = os.path.join(self.info_dir, self.args.std_mean_file)
        self.rgb_aug_dir = os.path.join(self.args.root_dir, self.args.rgb_aug_dir)
        if not os.path.exists(self.rgb_aug_dir): os.makedirs(self.rgb_aug_dir)
        self.label_aug_dir = os.path.join(self.args.root_dir, self.args.label_aug_dir)
        if not os.path.exists(self.label_aug_dir): os.makedirs(self.label_aug_dir)

    def poly2maskAug(self, inputs, filename='unname', title='viewer'):
        json_path = os.path.join(self.args.root_dir, self.args.ann_dir, filename + '.json')
        if not os.path.exists(json_path):
            print('polygon file does not exist ...')
            return 1

        im_in = inputs['rgb']
        height_in, width_in = im_in.shape[:2]

        json_file = open(json_path)
        ann_dict = json.load(json_file)
        json_file.close()
        label_in = polys2masks(ann_dict['Polygons'], im_in.shape[:2])

        # random select background,  angle, flip
        select_bg = random.sample(self.all_bg, k=self.num_aug)
        select_angle = random.sample(self.all_angle, k=self.num_aug)

        loc_flip = random.sample(range(self.num_aug), k=int(self.num_aug / 2))
        select_flip = np.zeros(shape=(self.num_aug, 1), dtype=np.uint8)
        select_flip[loc_flip] = 1
        select_flip = tuple(select_flip)

        if not os.path.exists(self.info_dir): os.makedirs(self.info_dir)
        if not os.path.exists(self.count_file): count=0
        else: count = int(np.load(self.count_file))
        if not os.path.exists(self.rgb_mean_file): rgb_mean=0
        else: rgb_mean = np.load(self.rgb_mean_file)
        if not os.path.exists(self.std_mean_file): std_mean=0
        else: std_mean = np.load(self.std_mean_file)
        mask = np.ones((height_in, width_in))
        for ind in range(self.num_aug):
            bg = select_bg[ind]
            angle = select_angle[ind]

            # crop to interesting region
            im_crop = np.copy(im_in)
            label_crop = np.copy(label_in)
            mask_crop = np.copy(mask)

            # rotate
            im_rot = rotate(im_crop, angle=angle, reshape=True, order=3)
            label_rot = rotate(label_crop, angle=angle, reshape=True, order=0)
            mask_rot = rotate(mask_crop, angle=angle, reshape=True, order=0)

            # selective flip and change color
            if bool(select_flip[ind]):
                im_rot = np.flip(im_rot, axis=0)
                label_rot = np.flip(label_rot, axis=0)
                mask_rot = np.flip(mask_rot, axis=0)

            # fix scales
            height_rot, width_rot, _ = im_rot.shape

            height_rat = self.args.out_size[0] / height_rot
            width_rat = self.args.out_size[1] / width_rot

            if (height_rat < 1) or (width_rat < 1):
                if height_rat < width_rat:
                    im_rot = cv2.resize(im_rot, None, fx=height_rat, fy=height_rat, interpolation=cv2.INTER_CUBIC)
                    label_rot = cv2.resize(label_rot, None, fx=height_rat, fy=height_rat,
                                           interpolation=cv2.INTER_NEAREST)
                    mask_rot = cv2.resize(mask_rot, None, fx=height_rat, fy=height_rat, interpolation=cv2.INTER_CUBIC)

                if width_rat <= height_rat:
                    im_rot = cv2.resize(im_rot, None, fx=width_rat, fy=width_rat, interpolation=cv2.INTER_CUBIC)
                    label_rot = cv2.resize(label_rot, None, fx=width_rat, fy=width_rat, interpolation=cv2.INTER_NEAREST)
                    mask_rot = cv2.resize(mask_rot, None, fx=width_rat, fy=width_rat, interpolation=cv2.INTER_CUBIC)

            height_rot, width_rot, _ = im_rot.shape

            # paste to background
            im_out = np.copy(bg)
            label_out = np.zeros(self.args.out_size, dtype=np.uint16)

            label_out[0:height_rot, 0:width_rot] = np.multiply(label_out[0:height_rot, 0:width_rot], 1 - mask_rot) + \
                                                   np.multiply(label_rot, mask_rot)
            for ii in range(3):
                im_out[0:height_rot, 0:width_rot, ii] = np.multiply(im_out[0:height_rot, 0:width_rot, ii],
                                                                    1 - mask_rot) + \
                                                        np.multiply(im_rot[:, :, ii], mask_rot)
            # save

            cv2.imwrite(os.path.join(self.rgb_aug_dir, str(count + self.args.aug_start_from) + '.png'), im_out)
            cv2.imwrite(os.path.join(self.label_aug_dir, str(count) + '.png'), label_out)

            rgb_mean = count/(count + 1)* rgb_mean + 1/(count+1)*np.mean(im_out.astype('float')/255, axis=(0, 1))
            std_mean = count/(count + 1)* std_mean + 1/(count+1)*np.std(im_out.astype('float') /255, axis=(0, 1))
            count += 1

        np.save(self.count_file, count)
        np.save(self.rgb_mean_file, rgb_mean)
        np.save(self.std_mean_file, std_mean)
        return 0







from common.utils.proc_utils import *
import torch
import torchvision
import torch.utils.data
import cv2
from ikea.model.classification_models import Net
# from model import classification_models as models
import os
import torchvision.transforms as transforms
import configparser
from scipy.ndimage import rotate


class Classifier():
    def __int__(self, conf_path):
        config = configparser.RawConfigParser()
        config.read(conf_path)

        rgb_mean = tuple([float(l.strip()) for l in config.get('data', 'rgb_mean').split(',')])
        rgb_std = tuple([float(l.strip()) for l in config.get('data', 'rgb_std').split(',')])
        self.transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize(rgb_mean, rgb_std)])
        self.classes = list([l.strip() for l in config.get('data', 'classes').split(',')])
        self.data_shape = tuple([int(l.strip()) for l in config.get('data', 'shape').split(',')])

        net_model = config.get('net', 'model')
        self.net = Net(net_model, input_shape=self.data_shape, num_classes=len(self.classes))

        net_path = config.get('net', 'path')
        self.net.load_state_dict(torch.load(net_path))

    def predict_array_single(self, im):

        im = fix_size(im, self.data_shape[:2])
        # cv2.imshow('scaled_im', im)
        im_tensor = self.transform(im)
        im_tensor = im_tensor.reshape((1,) + im_tensor.shape)
        outputs = self.net(im_tensor)
        _, predicted = torch.max(outputs, 1)
        probs = torch.softmax(outputs, 1)
        return self.classes[predicted[0]], probs[[0,predicted[0]]]

    def predict_list_array(self, im_list):

        num_im = len(im_list)
        inputs = None
        for im in im_list:
            im = fix_size(im, self.data_shape[:2])
            im_tensor = self.transform(im)
            im_tensor = im_tensor.reshape((1,) + im_tensor.shape)
            if inputs is None:
                inputs = im_tensor
            else:
                inputs = torch.cat((inputs, im_tensor), dim=0)

        outputs = self.net(inputs)
        _, predicted = torch.max(outputs, 1)

        out_classes = []
        for i in range(num_im):
            out_classes.append(self.classes[predicted[i]])

        return out_classes








import torch.nn as nn
import torch.nn.functional as F



class Net(nn.Module):

    def __init__(self):
        super(Net, self).__init__()
        # 3 input image channel, 6 output channels, 3x3 square convolution
        # kernel
        self.conv1 = nn.Conv2d(3, 6, 3)
        self.conv2 = nn.Conv2d(6, 16, 3)
        # an affine operation: y = Wx + b
        self.fc1 = nn.Linear(16 * 62 * 62, 120)  # 6*6 from image dimension
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 5)             # 5: number of classes  (without background)

    def forward(self, x):
        # Max pooling over a (2, 2) window
        x = F.max_pool2d(F.relu(self.conv1(x)), (2, 2))
        # If the size is a square you can only specify a single number
        x = F.max_pool2d(F.relu(self.conv2(x)), 2)
        x = x.view(-1, self.num_flat_features(x))
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

    def num_flat_features(self, x):
        size = x.size()[1:]  # all dimensions except the batch dimension
        num_features = 1
        for s in size:
            num_features *= s
        return num_features





'''Some helper functions for PyTorch, including:
    - get_mean_and_std: calculate the mean and std value of dataset.
    - msr_init: net parameter initialization.
    - progress_bar: progress bar mimic xlua.progress.
'''
import errno
import os
import sys
import time
import math

import torch.nn as nn
import torch.nn.init as init
from torch.autograd import Variable

__all__ = ['get_mean_and_std', 'init_params', 'mkdir_p', 'AverageMeter']


def get_mean_and_std(dataset):
    '''Compute the mean and std value of dataset.'''
    dataloader = trainloader = torch.utils.data.DataLoader(dataset, batch_size=1, shuffle=True, num_workers=2)

    mean = torch.zeros(3)
    std = torch.zeros(3)
    print('==> Computing mean and std..')
    for inputs, targets in dataloader:
        for i in range(3):
            mean[i] += inputs[:,i,:,:].mean()
            std[i] += inputs[:,i,:,:].std()
    mean.div_(len(dataset))
    std.div_(len(dataset))
    return mean, std

def init_params(net):
    '''Init layer parameters.'''
    for m in net.modules():
        if isinstance(m, nn.Conv2d):
            init.kaiming_normal(m.weight, mode='fan_out')
            if m.bias:
                init.constant(m.bias, 0)
        elif isinstance(m, nn.BatchNorm2d):
            init.constant(m.weight, 1)
            init.constant(m.bias, 0)
        elif isinstance(m, nn.Linear):
            init.normal(m.weight, std=1e-3)
            if m.bias:
                init.constant(m.bias, 0)

def mkdir_p(path):
    '''make dir if not exist'''
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

class AverageMeter(object):
    """Computes and stores the average and current value
       Imported from https://github.com/pytorch/examples/blob/master/imagenet/main.py#L247-L262
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


class TPS_SpatialTransformerNetwork(nn.Module):
    """ Rectification Network of RARE, namely TPS based STN """

    def __init__(self, F, I_size, I_r_size, I_channel_num=1):
        """ Based on RARE TPS
        input:
            batch_I: Batch Input Image [batch_size x I_channel_num x I_height x I_width]
            I_size : (height, width) of the input image I
            I_r_size : (height, width) of the rectified image I_r
            I_channel_num : the number of channels of the input image I
        output:
            batch_I_r: rectified image [batch_size x I_channel_num x I_r_height x I_r_width]
        """
        super(TPS_SpatialTransformerNetwork, self).__init__()
        self.F = F
        self.I_size = I_size
        self.I_r_size = I_r_size  # = (I_r_height, I_r_width)
        self.I_channel_num = I_channel_num
        self.LocalizationNetwork = LocalizationNetwork(self.F, self.I_channel_num)
        self.GridGenerator = GridGenerator(self.F, self.I_r_size)

    def forward(self, batch_I):
        batch_C_prime = self.LocalizationNetwork(batch_I)  # batch_size x K x 2
        build_P_prime = self.GridGenerator.build_P_prime(batch_C_prime)  # batch_size x n (= I_r_width x I_r_height) x 2
        build_P_prime_reshape = build_P_prime.reshape([build_P_prime.size(0), self.I_r_size[0], self.I_r_size[1], 2])
        batch_I_r = F.grid_sample(batch_I, build_P_prime_reshape, padding_mode='border')

        return batch_I_r


class LocalizationNetwork(nn.Module):
    """ Localization Network of RARE, which predicts C' (K x 2) from I (I_width x I_height) """

    def __init__(self, F, I_channel_num):
        super(LocalizationNetwork, self).__init__()
        self.F = F
        self.I_channel_num = I_channel_num
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels=self.I_channel_num, out_channels=64, kernel_size=3, stride=1, padding=1,
                      bias=False), nn.BatchNorm2d(64), nn.ReLU(True),
            nn.MaxPool2d(2, 2),  # batch_size x 64 x I_height/2 x I_width/2
            nn.Conv2d(64, 128, 3, 1, 1, bias=False), nn.BatchNorm2d(128), nn.ReLU(True),
            nn.MaxPool2d(2, 2),  # batch_size x 128 x I_height/4 x I_width/4
            nn.Conv2d(128, 256, 3, 1, 1, bias=False), nn.BatchNorm2d(256), nn.ReLU(True),
            nn.MaxPool2d(2, 2),  # batch_size x 256 x I_height/8 x I_width/8
            nn.Conv2d(256, 512, 3, 1, 1, bias=False), nn.BatchNorm2d(512), nn.ReLU(True),
            nn.AdaptiveAvgPool2d(1)  # batch_size x 512
        )

        self.localization_fc1 = nn.Sequential(nn.Linear(512, 256), nn.ReLU(True))
        self.localization_fc2 = nn.Linear(256, self.F * 2)

        # Init fc2 in LocalizationNetwork
        self.localization_fc2.weight.data.fill_(0)
        """ see RARE paper Fig. 6 (a) """
        ctrl_pts_x = np.linspace(-1.0, 1.0, int(F / 2))
        ctrl_pts_y_top = np.linspace(0.0, -1.0, num=int(F / 2))
        ctrl_pts_y_bottom = np.linspace(1.0, 0.0, num=int(F / 2))
        ctrl_pts_top = np.stack([ctrl_pts_x, ctrl_pts_y_top], axis=1)
        ctrl_pts_bottom = np.stack([ctrl_pts_x, ctrl_pts_y_bottom], axis=1)
        initial_bias = np.concatenate([ctrl_pts_top, ctrl_pts_bottom], axis=0)
        self.localization_fc2.bias.data = torch.from_numpy(initial_bias).float().view(-1)

    def forward(self, batch_I):
        """
        input:     batch_I : Batch Input Image [batch_size x I_channel_num x I_height x I_width]
        output:    batch_C_prime : Predicted coordinates of fiducial points for input batch [batch_size x F x 2]
        """
        batch_size = batch_I.size(0)
        features = self.conv(batch_I).view(batch_size, -1)
        batch_C_prime = self.localization_fc2(self.localization_fc1(features)).view(batch_size, self.F, 2)
        return batch_C_prime


class GridGenerator(nn.Module):
    """ Grid Generator of RARE, which produces P_prime by multipling T with P """

    def __init__(self, F, I_r_size):
        """ Generate P_hat and inv_delta_C for later """
        super(GridGenerator, self).__init__()
        self.eps = 1e-6
        self.I_r_height, self.I_r_width = I_r_size
        self.F = F
        self.C = self._build_C(self.F)  # F x 2
        self.P = self._build_P(self.I_r_width, self.I_r_height)
        ## for multi-gpu, you need register buffer
        self.register_buffer("inv_delta_C", torch.tensor(self._build_inv_delta_C(self.F, self.C)).float())  # F+3 x F+3
        self.register_buffer("P_hat", torch.tensor(self._build_P_hat(self.F, self.C, self.P)).float())  # n x F+3
        ## for fine-tuning with different image width, you may use below instead of self.register_buffer
        #self.inv_delta_C = torch.tensor(self._build_inv_delta_C(self.F, self.C)).float().cuda()  # F+3 x F+3
        #self.P_hat = torch.tensor(self._build_P_hat(self.F, self.C, self.P)).float().cuda()  # n x F+3

    def _build_C(self, F):
        """ Return coordinates of fiducial points in I_r; C """
        ctrl_pts_x = np.linspace(-1.0, 1.0, int(F / 2))
        ctrl_pts_y_top = -1 * np.ones(int(F / 2))
        ctrl_pts_y_bottom = np.ones(int(F / 2))
        ctrl_pts_top = np.stack([ctrl_pts_x, ctrl_pts_y_top], axis=1)
        ctrl_pts_bottom = np.stack([ctrl_pts_x, ctrl_pts_y_bottom], axis=1)
        C = np.concatenate([ctrl_pts_top, ctrl_pts_bottom], axis=0)
        return C  # F x 2

    def _build_inv_delta_C(self, F, C):
        """ Return inv_delta_C which is needed to calculate T """
        hat_C = np.zeros((F, F), dtype=float)  # F x F
        for i in range(0, F):
            for j in range(i, F):
                r = np.linalg.norm(C[i] - C[j])
                hat_C[i, j] = r
                hat_C[j, i] = r
        np.fill_diagonal(hat_C, 1)
        hat_C = (hat_C ** 2) * np.log(hat_C)
        # print(C.shape, hat_C.shape)
        delta_C = np.concatenate(  # F+3 x F+3
            [
                np.concatenate([np.ones((F, 1)), C, hat_C], axis=1),  # F x F+3
                np.concatenate([np.zeros((2, 3)), np.transpose(C)], axis=1),  # 2 x F+3
                np.concatenate([np.zeros((1, 3)), np.ones((1, F))], axis=1)  # 1 x F+3
            ],
            axis=0
        )
        inv_delta_C = np.linalg.inv(delta_C)
        return inv_delta_C  # F+3 x F+3

    def _build_P(self, I_r_width, I_r_height):
        I_r_grid_x = (np.arange(-I_r_width, I_r_width, 2) + 1.0) / I_r_width  # self.I_r_width
        I_r_grid_y = (np.arange(-I_r_height, I_r_height, 2) + 1.0) / I_r_height  # self.I_r_height
        P = np.stack(  # self.I_r_width x self.I_r_height x 2
            np.meshgrid(I_r_grid_x, I_r_grid_y),
            axis=2
        )
        return P.reshape([-1, 2])  # n (= self.I_r_width x self.I_r_height) x 2

    def _build_P_hat(self, F, C, P):
        n = P.shape[0]  # n (= self.I_r_width x self.I_r_height)
        P_tile = np.tile(np.expand_dims(P, axis=1), (1, F, 1))  # n x 2 -> n x 1 x 2 -> n x F x 2
        C_tile = np.expand_dims(C, axis=0)  # 1 x F x 2
        P_diff = P_tile - C_tile  # n x F x 2
        rbf_norm = np.linalg.norm(P_diff, ord=2, axis=2, keepdims=False)  # n x F
        rbf = np.multiply(np.square(rbf_norm), np.log(rbf_norm + self.eps))  # n x F
        P_hat = np.concatenate([np.ones((n, 1)), P, rbf], axis=1)
        return P_hat  # n x F+3

    def build_P_prime(self, batch_C_prime):
        """ Generate Grid from batch_C_prime [batch_size x F x 2] """
        batch_size = batch_C_prime.size(0)
        batch_inv_delta_C = self.inv_delta_C.repeat(batch_size, 1, 1)
        batch_P_hat = self.P_hat.repeat(batch_size, 1, 1)
        batch_C_prime_with_zeros = torch.cat((batch_C_prime, torch.zeros(
            batch_size, 3, 2).float().to(device)), dim=1)  # batch_size x F+3 x 2
        batch_T = torch.bmm(batch_inv_delta_C, batch_C_prime_with_zeros)  # batch_size x F+3 x 2
        batch_P_prime = torch.bmm(batch_P_hat, batch_T)  # batch_size x n x 2
        return batch_P_prime  # batch_size x n x 2



from libs.basic.basic_sequence import BasSeqImDetectCallback
from libs.basic.basic_imscanning_ray import EdgePairFinder
from libs.import_basic_utils import *
from libs.basic.basic_objects import BasDoLabel, BasJson
from grasp_detection.processing.grip_detection_v6 import GripDetector
import cv2
from time import time
import os



class LabelGripEdge(BasSeqImDetectCallback):
    def get_model(self):
        super().get_model()
        # self.model = EdgePairFinder(cfg_path=self.cfg_path)
        # self.detector = self.model.find_grip_candidates_from_edges
        self.model = GripDetector(cfg_path=self.cfg_path)
        self.detector = self.model.find_grip_pose_from_edges_v2
        self.labeler = BasDoLabel(cfg_path=self.cfg_path)


    def detect(self):
        self.Grips = self.detector(rgbd=self.rgbd)
        # self.grips = self.Grips.grips
        self.grips = self.Grips.get_top_higher_grips(score=0.2)
        # self.grips = self.detector(rgbd=self.rgbd)

    def process_single(self):
        super().process_single()
        t = time()

        self.detect()
        if self.grips is None:
            print('Grip pairing found no grip ...')
            return self.rgbd, None

        num_grip = len(self.grips)
        print('Found %d grip candidates in %0.3f s' % (num_grip, time() - t))

        self.rgbd.show(args=self.args, grips=GRIPS(grips=self.grips))
        print('ESC: next image, ANY KEY: continue')
        if cv2.waitKey() != ESC_KEY:
            self.labeled_grips = []
            for i in range(num_grip):
                ProcUtils().clscr()
                print('Grip %d out of %d ...' %(i,num_grip))
                self.rgbd.show(args=self.args, grips=GRIPS(grips=[self.grips[i]], disp_colors=[(0, 255, 0)]))
                print('ENTER: skip this, ANY KEY: continue')
                if cv2.waitKey() == ENTER_KEY: continue

                label_ret = self.labeler.get_key_label()


                if label_ret is not None:
                    self.labeled_grips.append(self.grips[i])
                    self.rgbd.show(args=self.args,
                                   grips=GRIPS(grips=[self.labeled_grips[-1]],
                                               disp_colors=[self.labeler.disp_colors[-1]]))

                print('ESC: next image, ANY KEY: continue')
                if cv2.waitKey() == ESC_KEY: break

            Grips = None
            if len(self.labeled_grips)>0: Grips = GRIPS(grips=self.labeled_grips, disp_colors=self.labeler.disp_colors)
            self.rgbd.show(args=self.args, grips=Grips)

            # save
            top, left, _, _ =  self.rgbd.roi
            for i in range(len(self.labeled_grips)): self.labeled_grips[i].shift(dx=-left, dy=-top)
            polys = [grip.pts for grip in self.labeled_grips]
            jsonname, _ = os.path.splitext(self.filename)
            BasJson().save_poly_json(jsonname + '.json', self.labeler.get_im_info(self.rgbd.rgb, self.filename),
                                     self.labeler.class_list, polys, [self.args.root_dir, self.args.ann_poly])
            ArrayUtils().save_array(self.rgbd.crop_rgb()[:,:,::-1], [self.args.root_dir, self.args.ann_rgb], self.filename)
            ArrayUtils().save_array(self.rgbd.crop_depth(), [self.args.root_dir, self.args.ann_depth], self.filename)

            # move file
            self.move_processed_rgbd()






if __name__=='__main__':
    LabelGripEdge(cfg_path='configs/grasp_detection/grip_v3.cfg').run()


from PIL import Image
import os
import os.path
import numpy as np
import sys

if sys.version_info[0] == 2:
    import cPickle as pickle
else:
    import pickle

from torchvision.datasets.vision import VisionDataset
from torchvision.datasets.utils import download_url, check_integrity


class GripCifar10(VisionDataset):

    base_folder = 'cifar10'
    url = None
    filename = None
    tgz_md5 = None
    train_list = [['data_batch', None]]

    test_list = [['test_batch', None]]
    meta = {
        'filename': 'batches.meta',
        'key': 'label_names',
        'md5': None,
    }

    # db_mean = (0.61799215, 0.29761591, 0.37573633, 0.52014014)#(0.60075633,0.30408495, 0.37778071, 0.22482509)
    # db_std = (0.10109334, 0.1636474, 0.14329318, 0.0625928)#(0.11525005,0.2022998,  0.17382977, 0.09038225)

    # db_mean = (0.60754877, 0.38245198, 0.44624274, 0.43154834)
    # db_std = (0.14569339, 0.18942552, 0.17363705, 0.02090128)

    # db_mean = (0.42988802, 0.42988802, 0.42988802)    #grip_evaluator_depth_20200327
    # db_std = (0.02043479, 0.02043479, 0.02043479)

    # db_mean = (0.60747145, 0.3981848,  0.45978323)    # grip_evaluator_rgb_20200327
    # db_std = (0.14146864, 0.18522996, 0.17041358)

    # db_mean = (0.55005549, 0.55005549, 0.55005549)      # grip_evaluator_depth_20200329
    # db_std = (0.03288065, 0.03288065, 0.03288065)

    # db_mean = (0.55184516, 0.55184516, 0.55184516)      # grip_evaluator_depth_20200408
    # db_std = (0.03759594, 0.03759594, 0.03759594)

    # db_mean = (0.54323941, 0.54323941, 0.54323941)  # grip_evaluator_depth_20200409
    # db_std = (0.04577378, 0.04577378, 0.04577378)

    # db_mean = (0.54602665, 0.54602665, 0.54602665)  # grip_evaluator_depth_20200423
    # db_std = (0.03714684, 0.03714684, 0.03714684)

    # db_mean = (0.35978138, 0.41708053, 0.5188343,  0.53351979)  # grip_evaluator_depth_20200424
    # db_std = (0.11914077, 0.10032271, 0.12174645, 0.04099747)

    # db_mean = (0.33610585, 0.38967911, 0.49009963, 0.52970528)  # grip_evaluator_depth_20200427
    # db_std = (0.12106411, 0.10579008, 0.12965363, 0.04399332)

    # db_mean = (0.42963929 0.44455258 0.42168464 0.50111799)  # 20200512, or use 202200427
    # db_std = (0.145406   0.137699   0.1424108  0.03611386)

    # db_mean = (0.41280124, 0.42837454, 0.42626716, 0.31356105)  # 20200514
    # db_std = (0.15362443, 0.15150588, 0.15615994, 0.03650546)

    # db_mean = (0.36361247, 0.40355713, 0.46720627, 0.45218567)  # 20200518
    # db_std = (0.13274179, 0.12218594, 0.13916006, 0.04130782)

    # db_mean = (0.76513104, 0.74994571, 0.74697943, 0.4860831)  # 20200610
    # db_std = (0.17024282, 0.1789022,  0.1872297,  0.09504798)

    db_mean = (0.41828456, 0.40237707, 0.41067797, 0.24735339, 0.1663269,  0.85723261)  # 20200624
    db_std = (0.14510322, 0.14341601, 0.14637822, 0.25664706, 0.21761108, 0.16026509)

    def __init__(self, root=None, train=True,
                 transform=None, target_transform=None,
                 download=False, im_shape=(32, 32, 3), data=None, indexing=False):

        super(GripCifar10, self).__init__(root)
        self.transform = transform
        self.target_transform = target_transform

        self.indexing = indexing

        self.train = train  # training set or test set
        if download:
            self.download()

        if data is None:
            if not self._check_integrity():
                raise RuntimeError('Dataset not found or corrupted.' +
                                   ' You can use download=True to download it')

        if self.train:
            downloaded_list = self.train_list
        else:
            downloaded_list = self.test_list

        self.data = []
        self.targets = []
        if data is not None: self.targets = [0] * len(data)
        if data is None:
            # now load the picked numpy arrays
            for file_name, checksum in downloaded_list:
                file_path = os.path.join(self.root, self.base_folder, file_name)
                with open(file_path, 'rb') as f:
                    if sys.version_info[0] == 2:
                        entry = pickle.load(f)
                    else:
                        # entry = pickle.load(f, encoding='latin1')
                        entry = pickle.load(f)
                    self.data.append(entry['data'])
                    if 'labels' in entry:
                        self.targets.extend(entry['labels'])
                    else:
                        self.targets.extend(entry['fine_labels'])

            self._load_meta()
            self.data = np.vstack(self.data).reshape(-1, im_shape[2], im_shape[0], im_shape[1])
            self.data = self.data.transpose((0, 2, 3, 1))  # convert to HWC
        else:
            self.data = np.vstack([p.reshape((1,) + p.shape) for p in data])


    def array2cifar(self, anArray):
        h, w, ch = anArray.shape
        out = [anArray[:, :, i].reshape(1, h * w) for i in range(ch)]
        return np.hstack(out)

    def _load_meta(self):
        path = os.path.join(self.root, self.base_folder, self.meta['filename'])
        if not check_integrity(path, self.meta['md5']):
            raise RuntimeError('Dataset metadata file not found or corrupted.' +
                               ' You can use download=True to download it')
        with open(path, 'rb') as infile:
            if sys.version_info[0] == 2:
                data = pickle.load(infile)
            else:
                data = pickle.load(infile, encoding='latin1')
            self.classes = data[self.meta['key']]
        self.class_to_idx = {_class: i for i, _class in enumerate(self.classes)}

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        # img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.indexing:
            return img, target, index
        else:
            return img, target

    def __len__(self):
        return len(self.data)

    def _check_integrity(self):
        root = self.root
        for fentry in (self.train_list + self.test_list):
            filename, md5 = fentry[0], fentry[1]
            fpath = os.path.join(root, self.base_folder, filename)
            if not check_integrity(fpath, md5):
                return False
        return True

    def download(self):
        import tarfile

        if self._check_integrity():
            print('Files already downloaded and verified')
            return

        download_url(self.url, self.root, self.filename, self.tgz_md5)

        # extract file
        with tarfile.open(os.path.join(self.root, self.filename), "r:gz") as tar:
            tar.extractall(path=self.root)

    def extra_repr(self):
        return "Split: {}".format("Train" if self.train is True else "Test")




'''VGG for CIFAR10. FC layers are removed.
(c) YANG, Wei 
'''
import torch.nn as nn
import torch.utils.model_zoo as model_zoo
import math


__all__ = [
    'VGG', 'vgg11', 'vgg11_bn', 'vgg13', 'vgg13_bn', 'vgg16', 'vgg16_bn',
    'vgg19_bn', 'vgg19',
]


model_urls = {
    'vgg11': 'https://download.pytorch.org/models/vgg11-bbd30ac9.pth',
    'vgg13': 'https://download.pytorch.org/models/vgg13-c768596a.pth',
    'vgg16': 'https://download.pytorch.org/models/vgg16-397923af.pth',
    'vgg19': 'https://download.pytorch.org/models/vgg19-dcbb9e9d.pth',
}


class VGG(nn.Module):

    def __init__(self, features, num_classes=1000):
        super(VGG, self).__init__()
        self.features = features
        self.classifier = nn.Linear(512, num_classes)
        self._initialize_weights()

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()


def make_layers(cfg, batch_norm=False):
    layers = []
    in_channels = 3
    for v in cfg:
        if v == 'M':
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
        else:
            conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
            if batch_norm:
                layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
            else:
                layers += [conv2d, nn.ReLU(inplace=True)]
            in_channels = v
    return nn.Sequential(*layers)


cfg = {
    'A': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'B': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'D': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'E': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}


def vgg11(**kwargs):
    """VGG 11-layer model (configuration "A")

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = VGG(make_layers(cfg['A']), **kwargs)
    return model


def vgg11_bn(**kwargs):
    """VGG 11-layer model (configuration "A") with batch normalization"""
    model = VGG(make_layers(cfg['A'], batch_norm=True), **kwargs)
    return model


def vgg13(**kwargs):
    """VGG 13-layer model (configuration "B")

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = VGG(make_layers(cfg['B']), **kwargs)
    return model


def vgg13_bn(**kwargs):
    """VGG 13-layer model (configuration "B") with batch normalization"""
    model = VGG(make_layers(cfg['B'], batch_norm=True), **kwargs)
    return model


def vgg16(**kwargs):
    """VGG 16-layer model (configuration "D")

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = VGG(make_layers(cfg['D']), **kwargs)
    return model


def vgg16_bn(**kwargs):
    """VGG 16-layer model (configuration "D") with batch normalization"""
    model = VGG(make_layers(cfg['D'], batch_norm=True), **kwargs)
    return model


def vgg19(**kwargs):
    """VGG 19-layer model (configuration "E")

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = VGG(make_layers(cfg['E']), **kwargs)
    return model


def vgg19_bn(**kwargs):
    """VGG 19-layer model (configuration 'E') with batch normalization"""
    model = VGG(make_layers(cfg['E'], batch_norm=True), **kwargs)
    return model



from libs.dataset.polygon import *
from libs.dataset.cifar10 import Cifar10Maker, CombineCifar10Data
from libs.dataset.polygon import OnMousePolygonDrawer, PolygonShower, PolygonAnalyzer
from libs.import_basic_utils import *
import json
import numpy as np
import math

class GripCifar10DMaker(Cifar10Maker):
    def load_params(self, args):
        super().load_params(args=args)
        self.ann_poly = os.path.join(self.args.root_dir, self.args.ann_poly)
        assert ProcUtils().isexists(self.ann_poly)

    def process_single(self):
    # def process_single(self):
        # read polygon from json
        json_path = os.path.join(self.ann_poly,self.filename+'.json')
        assert ProcUtils().isexists(json_path)
        json_file = open(json_path)
        ann_dict = json.load(json_file)
        json_file.close()
        classes = ann_dict['Classes']
        polygons = ann_dict['Polygons']

        # pad
        h,w = self.rgbd.height, self.rgbd.width
        diag = int(np.linalg.norm((h,w)))
        top, left = int((diag-h)/2), int((diag-w)/2)
        bottom, right = diag-h-top, diag-w-left
        center = (int(diag/2), int(diag/2))

        rgbd_pad =  self.rgbd.pad(top=top, left=left, bottom=bottom, right=right)

        count = 0
        for cls, poly in zip(classes, polygons):
            pt1, pt2, = poly[:2]
            x1, y1 = pt1
            x2, y2 = pt2
            if self.args.show_steps: self.rgbd.show(title='rgbd', args=self.args, grips=GRIPS(grips=[GRIP(pts=[(x1,y1),(x2,y2)])]))

            # compute angle
            dx, dy = pt1[0] - pt2[0], pt1[1] - pt2[1]
            angle = math.atan(dy/(dx+0.000001)) * 180 /math.pi

            # totate
            rgbd_rot = rgbd_pad.rotate(angle=angle)

            # new pt (y1=y2)
            x1, y1 = x1+left, y1+top
            x1,y1 = ProcUtils().rotateXY(x1,y1, -angle, org=center)
            x2, y2 = x2 + left, y2 + top
            x2, y2 = ProcUtils().rotateXY(x2, y2, -angle, org=center)
            if self.args.show_steps:
                rgbd_rot.show(title='rgbd_rot',args=self.args, grips=GRIPS(grips=[GRIP(pts=[(x1,y1),(x2,y2)])]))

            xmin, xmax = min(x1, x2), max(x1, x2)
            # crop
            grip_hs, grip_w_margins =self.args.train_grip_hs, self.args.train_grip_w_margins
            # if cls=='ungrip' and self.filename not in self.args.feedback_images:
            #     grip_hs, grip_w_margins = self.args.ungrip_hs, self.args.ungrip_w_margins

            for grip_h in grip_hs:
                for grip_w_margin in grip_w_margins:
                    count += 1
                    r = int(grip_h/2)
                    rgbd_crop=rgbd_rot.crop(left=xmin-grip_w_margin, top=y1-r, right=xmax+1+grip_w_margin, bottom=y1+r+1)

                    if self.args.show_steps:
                        rgbd_crop.show(title='rgbd_crop',args=self.args)

                    self.cls_ind = self.args.classes.index(cls)
                    super().process_single(rgbd=rgbd_crop)
        print('>>> %d samples ...' % count)

class OnMouseDrawGrip(OnMousePolygonDrawer):

    def show_draws(self, rgbd):
        grips = [GRIP(pts=poly[:2]) for poly in self.polygons]
        Grips = GRIPS(grips=grips, disp_colors=self.disp_colors)
        self.im = rgbd.disp(args=self.args, grips=Grips)


if __name__ =='__main__':
    cfg_path = 'configs/grasp_detection/grip_net.cfg'

    # OnMouseDrawGrip(cfg_path=cfg_path).run()   #1
    # PolygonShower(cfg_path=cfg_path).run() #2
    # GripCifar10DMaker(cfg_path=cfg_path).run() #3


    # PolygonAnalyzer(cfg_path=cfg_path).run()
    CombineCifar10Data(cfg_path='configs/grasp_detection/grip_net.cfg').run()








import cv2
import numpy as np

def nms(boxes, scores, score_thresh=0.5, nms_thresh=0.7):
    # boxes = [[left, top, right, bottom]]
    boxes_array = np.array(boxes)               # convert to x,y,w,h
    boxes_array[:, 2] -= boxes_array[:,0]
    boxes_array[:, 3] -= boxes_array[:, 1]
    bboxes = boxes_array.tolist()

    idxs = cv2.dnn.NMSBoxes(bboxes, scores,
                            score_threshold=score_thresh, nms_threshold=nms_thresh).flatten()
    return [boxes[i] for i in idxs], [scores[i] for i in idxs]


from libs.dataset.polygon import *
from libs.dataset.cifar10 import Cifar10Maker, CombineCifar10Data, CifarDepth2Norm
from libs.dataset.polygon import OnMousePolygonDrawer, PolygonShower, PolygonAnalyzer
from libs.import_basic_utils import *
import json
import numpy as np
import math

class SuctionCifar10DMaker(Cifar10Maker):

    def process_this(self):
        self.mask_path = self.rgbd_path[0].replace('.png', '_mask.png')
        if not os.path.exists(self.mask_path):
            print('>>> %s does not exist >>> Skip' %self.mask_path)
            return False
        return super().process_this()

    def process_single(self):
        # read mask
        mask = cv2.imread(self.mask_path, cv2.IMREAD_GRAYSCALE)
        mask_crop = ArrayUtils().rect_crop(mask, self.args.workspace.rect)

        rgbd_crop = self.rgbd.crop()




        # bound_mask = np.logical_and(50<mask_crop, mask_crop<200).astype('uint8')
        # bound_mask  =cv2.dilate(bound_mask, np.ones((3,3), 'uint8'))
        # suc_mask = np.logical_and(mask_crop>200,1-bound_mask).astype('uint8')
        # bg_mask = 1-(bound_mask+suc_mask)

        suc_mask_org = (mask_crop > 150).astype('uint8')
        suc_mask = cv2.erode(suc_mask_org,np.ones(self.args.suction_size, 'uint8'))
        bound_mask = cv2.dilate(suc_mask_org,np.ones(self.args.suction_size, 'uint8')) - suc_mask
        bg_mask = 1 - (bound_mask + suc_mask)


        locs_suc = np.where(suc_mask)
        locs_bound = np.where(bound_mask)
        locs_bg = np.where(bg_mask)

        num_suc_px = len(locs_suc[0])
        num_bound_px = len(locs_bound[0])
        num_bg_px = len(locs_bg[0])

        if self.args.show_steps:
            # cv2.imshow('mask_crop', mask_crop)
            # cv2.imshow('suc_mask', 255*suc_mask)
            # cv2.imshow('bg_mask', 255*bg_mask)
            # cv2.imshow('bound_mask', 255*bound_mask)
            rgbd_crop.show(args=self.args, mask=suc_mask)

        suc_mask = MaskUtils().sparse_mask(suc_mask, stride=round(math.sqrt(num_suc_px/self.args.num_roi)))
        bound_mask = MaskUtils().sparse_mask(bound_mask, stride=round(math.sqrt(num_bound_px / self.args.num_roi)))
        bg_mask = MaskUtils().sparse_mask(bg_mask, stride=round(math.sqrt(num_bg_px / self.args.num_roi)))

        locs_suc = np.where(suc_mask)
        locs_bound = np.where(bound_mask)
        locs_bg = np.where(bg_mask)

        num_suc_px = len(locs_suc[0])
        num_bound_px = len(locs_bound[0])
        num_bg_px = len(locs_bg[0])

        count = 0
        for pad in self.args.roi_pads:
            xp, yp = pad
            sleft, stop = xp//2, yp//2
            sright, sbottom = xp-sleft, yp-stop
            # for mmask, cls in zip([suc_mask, bound_mask, bg_mask], ['suction', 'bound', 'bg']):
            for mmask, cls in zip([suc_mask, bound_mask], ['suction', 'bound']):
                Y, X = np.where(mmask)
                self.cls_ind = self.args.classes.index(cls)

                for x, y in zip(X, Y):
                    rgbd_roi = rgbd_crop.crop(left=x-sleft, right=x+sright, top=y-stop, bottom=y+sbottom)
                    if self.args.show_steps: rgbd_crop.show(args=self.args, pt=(x,y), title='location')

                    # rgbd_crop.show(args=self.args, pt=(x,y), mask=suc_mask)
                    # rgbd_roi.show(args=self.args, title='roi')
                    # cv2.waitKey()
                    super().process_single(rgbd=rgbd_roi)
                    count+=1

        print('>>> %d samples ...' % count)



if __name__ =='__main__':
    cfg_path = 'configs/grasp_detection/suction_net.cfg'

    SuctionCifar10DMaker(cfg_path=cfg_path).run()
    # PolygonShower(cfg_path=cfg_path).run()
    # OnMouseDrawGrip(cfg_path=cfg_path).run()
    # PolygonAnalyzer(cfg_path=cfg_path).run()

    # CombineCifar10Data(cfg_path=cfg_path).run()
    # CifarDepth2Norm(cfg_path=cfg_path).run()







'''AlexNet for CIFAR10. FC layers are removed. Paddings are adjusted.
Without BN, the start learning rate should be 0.01
(c) YANG, Wei 
'''
import torch.nn as nn


__all__ = ['alexnet']


class AlexNet(nn.Module):

    def __init__(self, num_classes=10):
        super(AlexNet, self).__init__()
        self.features = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=11, stride=4, padding=5),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(64, 192, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )
        self.classifier = nn.Linear(256, num_classes)

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x


def alexnet(**kwargs):
    r"""AlexNet model architecture from the
    `"One weird trick..." <https://arxiv.org/abs/1404.5997>`_ paper.
    """
    model = AlexNet(**kwargs)
    return model



import torch
import torch.nn as nn
import torch.nn.functional as F
import math


__all__ = ['densenet']


from torch.autograd import Variable

class Bottleneck(nn.Module):
    def __init__(self, inplanes, expansion=4, growthRate=12, dropRate=0):
        super(Bottleneck, self).__init__()
        planes = expansion * growthRate
        self.bn1 = nn.BatchNorm2d(inplanes)
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, growthRate, kernel_size=3, 
                               padding=1, bias=False)
        self.relu = nn.ReLU(inplace=True)
        self.dropRate = dropRate

    def forward(self, x):
        out = self.bn1(x)
        out = self.relu(out)
        out = self.conv1(out)
        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv2(out)
        if self.dropRate > 0:
            out = F.dropout(out, p=self.dropRate, training=self.training)

        out = torch.cat((x, out), 1)

        return out


class BasicBlock(nn.Module):
    def __init__(self, inplanes, expansion=1, growthRate=12, dropRate=0):
        super(BasicBlock, self).__init__()
        planes = expansion * growthRate
        self.bn1 = nn.BatchNorm2d(inplanes)
        self.conv1 = nn.Conv2d(inplanes, growthRate, kernel_size=3, 
                               padding=1, bias=False)
        self.relu = nn.ReLU(inplace=True)
        self.dropRate = dropRate

    def forward(self, x):
        out = self.bn1(x)
        out = self.relu(out)
        out = self.conv1(out)
        if self.dropRate > 0:
            out = F.dropout(out, p=self.dropRate, training=self.training)

        out = torch.cat((x, out), 1)

        return out


class Transition(nn.Module):
    def __init__(self, inplanes, outplanes):
        super(Transition, self).__init__()
        self.bn1 = nn.BatchNorm2d(inplanes)
        self.conv1 = nn.Conv2d(inplanes, outplanes, kernel_size=1,
                               bias=False)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        out = self.bn1(x)
        out = self.relu(out)
        out = self.conv1(out)
        out = F.avg_pool2d(out, 2)
        return out


class DenseNet(nn.Module):

    def __init__(self, depth=22, block=Bottleneck, 
        dropRate=0, num_classes=10, growthRate=12, compressionRate=2):
        super(DenseNet, self).__init__()

        assert (depth - 4) % 3 == 0, 'depth should be 3n+4'
        n = (depth - 4) / 3 if block == BasicBlock else (depth - 4) // 6

        self.growthRate = growthRate
        self.dropRate = dropRate

        # self.inplanes is a global variable used across multiple
        # helper functions
        self.inplanes = growthRate * 2 
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=3, padding=1,
                               bias=False)
        self.dense1 = self._make_denseblock(block, n)
        self.trans1 = self._make_transition(compressionRate)
        self.dense2 = self._make_denseblock(block, n)
        self.trans2 = self._make_transition(compressionRate)
        self.dense3 = self._make_denseblock(block, n)
        self.bn = nn.BatchNorm2d(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.avgpool = nn.AvgPool2d(8)
        self.fc = nn.Linear(self.inplanes, num_classes)

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_denseblock(self, block, blocks):
        layers = []
        for i in range(blocks):
            # Currently we fix the expansion ratio as the default value
            layers.append(block(self.inplanes, growthRate=self.growthRate, dropRate=self.dropRate))
            self.inplanes += self.growthRate

        return nn.Sequential(*layers)

    def _make_transition(self, compressionRate):
        inplanes = self.inplanes
        outplanes = int(math.floor(self.inplanes // compressionRate))
        self.inplanes = outplanes
        return Transition(inplanes, outplanes)


    def forward(self, x):
        x = self.conv1(x)

        x = self.trans1(self.dense1(x)) 
        x = self.trans2(self.dense2(x)) 
        x = self.dense3(x)
        x = self.bn(x)
        x = self.relu(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x


def densenet(**kwargs):
    """
    Constructs a ResNet model.
    """
    return DenseNet(**kwargs)


from mmdet.apis import init_detector, inference_detector, show_result
import mmcv
import numpy as np
import pycocotools.mask as maskUtils
from time import time
import configparser
import cv2
from ttcv.import_basic_utils import *
from ttcv.basic.basic_objects import BasDetectObj

class FasterRCNNDetector():
    def get_model(self, config_file, checkpoint_file):
        self.model = init_detector(config_file, checkpoint_file, device='cuda:0')

    def predict_array(self, anArray, need_rot=False, test_shape=None):
        h,w = anArray.shape[:2]
        if need_rot: im = np.rot90(anArray, k=-1)           # 90 rot
        else: im = anArray

        need_reshape = (test_shape is not None)
        if need_reshape:
            h1, w1 = im.shape[:2]
            im = cv2.resize(im, test_shape, interpolation=cv2.INTER_CUBIC)
            h2, w2 = im.shape[:2]
            fx, fy = w1/w2, h1/h2       # restore ratios

        boxes_list=inference_detector(self.model, im)

        if need_reshape:                # unscale
            for boxes in boxes_list:
                for box in boxes:
                    box[[0,2]] *= fx
                    box[[1, 3]] *= fy


        if need_rot:
            new_boxes_list = []
            for boxes in boxes_list:
                if len(boxes) ==0:
                    new_boxes_list.append(boxes)
                    continue
                new_boxes = []
                for box in boxes:
                    box = box[[1,2,3,0,4]]
                    box[1] = h-box[1]
                    box[3] = h - box[3]
                    new_boxes.append(box)
                new_boxes_list.append(new_boxes)
        else: new_boxes_list = boxes_list

        return new_boxes_list














from processing.proc_utils import fix_size
import numpy as np
from keras.models import load_model
import cv2
from glob import glob
import os, random


def make_test_pairs(test_im_in, support_ims_array, test_scale=(105, 105)):
    num_support_im = support_ims_array.shape[0]

    test_im = fix_size(test_im_in, test_scale)
    test_im = test_im.reshape((1, ) + test_im.shape + (1,))
    test_ims_array = np.tile(test_im, (num_support_im,1,1,1))

    return [test_ims_array, support_ims_array]


def make_support_set( support_set_dir, test_scale=(105, 105)):
    all_support_subset_dir = glob(os.path.join(support_set_dir, '*'))
    num_chr = len(all_support_subset_dir)
    support_ims = []
    for support_subset in all_support_subset_dir:
        all_support_im = glob(os.path.join(support_subset, '*'))
        num_support_im = len(all_support_im)
        support_im_ind = random.randint(0, num_support_im - 1)
        support_im = fix_size(cv2.imread(all_support_im[support_im_ind], 0), test_scale)

        support_ims.append(support_im.reshape((1,) + support_im.shape))

    support_ims_array = np.vstack(support_ims)
    support_ims_array = support_ims_array.reshape(support_ims_array.shape + (1,))

    return support_ims_array


class SiameseDetector():
    def __int__(self, support_set_dir, test_scale):
        self.model = load_model('../detector/siamese_keras/weights.h5')
        self.test_scale = test_scale
        self.support_set = make_support_set(support_set_dir, self.test_scale)


    def predict_array(self, im):
        print('---------------------------->>>>>>>>>> Siamese detection')
        pairs = make_test_pairs(im, self.support_set, self.test_scale)

        p = self.model.predict(pairs)
        max_ind = int(np.argmax(np.array(p)))
        target = self.support_set[max_ind][:, :, 0]
        target = fix_size(target, im.shape)

        return target, p

    def show_detect_results(self, im, target):
        out = np.concatenate((im, target))
        return out










from __future__ import absolute_import

"""The models subpackage contains definitions for the following model for CIFAR10/CIFAR100
architectures:

-  `AlexNet`_
-  `VGG`_
-  `ResNet`_
-  `SqueezeNet`_
-  `DenseNet`_

You can construct a model with random weights by calling its constructor:

.. code:: python

    import torchvision.models as models
    resnet18 = models.resnet18()
    alexnet = models.alexnet()
    squeezenet = models.squeezenet1_0()
    densenet = models.densenet_161()

We provide pre-trained models for the ResNet variants and AlexNet, using the
PyTorch :mod:`torch.utils.model_zoo`. These can  constructed by passing
``pretrained=True``:

.. code:: python

    import torchvision.models as models
    resnet18 = models.resnet18(pretrained=True)
    alexnet = models.alexnet(pretrained=True)

ImageNet 1-crop error rates (224x224)

======================== =============   =============
Network                  Top-1 error     Top-5 error
======================== =============   =============
ResNet-18                30.24           10.92
ResNet-34                26.70           8.58
ResNet-50                23.85           7.13
ResNet-101               22.63           6.44
ResNet-152               21.69           5.94
Inception v3             22.55           6.44
AlexNet                  43.45           20.91
VGG-11                   30.98           11.37
VGG-13                   30.07           10.75
VGG-16                   28.41           9.62
VGG-19                   27.62           9.12
SqueezeNet 1.0           41.90           19.58
SqueezeNet 1.1           41.81           19.38
Densenet-121             25.35           7.83
Densenet-169             24.00           7.00
Densenet-201             22.80           6.43
Densenet-161             22.35           6.20
======================== =============   =============


.. _AlexNet: https://arxiv.org/abs/1404.5997
.. _VGG: https://arxiv.org/abs/1409.1556
.. _ResNet: https://arxiv.org/abs/1512.03385
.. _SqueezeNet: https://arxiv.org/abs/1602.07360
.. _DenseNet: https://arxiv.org/abs/1608.06993
"""

from .alexnet import *
from .vgg import *
from .resnet import *
from .resnext import *
from .wrn import *
from .preresnet import *
from .densenet import *



from KAI.policy.sequence_accumulate import  SeqAccumulator
from KAI.basic.basic_objects import BasJson, BasPoly
import os
from KAI.utils.proc_utils import isexists
import json
import cv2

class DotaMaker(SeqAccumulator, BasJson, BasPoly):
    def load_params(self, args):
        super().load_params(args=args)
        self.ann_dir = os.path.join(self.args.root_dir, self.args.ann_dir)
        assert isexists(self.ann_dir)

    def process(self, rgbd, filename='unnamed'):
        # read polygon from json
        filename, _ = os.path.splitext(filename)
        json_path = os.path.join(self.ann_dir, filename + '.json')
        ann_dict = self.read_json(json_path=json_path)

        classes = ann_dict['Classes']
        polygons = ann_dict['Polygons']

        if self.args.show_steps:
            out = self.draw_polys_from_json(json_path=json_path, im=rgbd.rgb)
            cv2.imshow('out', out)



        if self.args.show_steps:
            if cv2.waitKey()==27: exit()

if __name__=='__main__':
    DotaMaker(cfg_path='configs/dataset/make_grip_dota_db.cfg').run()









import torch.nn as nn


class BidirectionalLSTM(nn.Module):

    def __init__(self, input_size, hidden_size, output_size):
        super(BidirectionalLSTM, self).__init__()
        self.rnn = nn.LSTM(input_size, hidden_size, bidirectional=True, batch_first=True)
        self.linear = nn.Linear(hidden_size * 2, output_size)

    def forward(self, input):
        """
        input : visual feature [batch_size x T x input_size]
        output : contextual feature [batch_size x T x output_size]
        """
        self.rnn.flatten_parameters()
        recurrent, _ = self.rnn(input)  # batch_size x T x input_size -> batch_size x T x (2*hidden_size)
        output = self.linear(recurrent)  # batch_size x T x output_size
        return output



from .grip_cifar10_dber import GripCifar10
from .suction_cifar10_dber import SuctionCifar10
from torchvision.datasets import *
import torchvision


__all__ = ('GripCifar10','SuctionCifar10') + torchvision.datasets.__all__

def get_mean_std(name):
    assert name in __all__ + torchvision.datasets.__all__
    if name in torchvision.datasets.__all__:
        db_mean = (0.4914, 0.4822, 0.4465)
        db_std = (0.2023, 0.1994, 0.2010)
    else:
        db_mean = eval(name + '.db_mean')
        db_std = eval(name + '.db_std')
    return db_mean, db_std

def Dataset(name, **kwargs):
    assert name in __all__ + torchvision.datasets.__all__
    if name in torchvision.datasets.__all__: del kwargs['im_shape']
    return eval(name)(**kwargs)